# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from typing import Any, Generic, TypeVar, cast

import pytest
import simpy as SIM

import upstage_des.api as UP
from upstage_des.geography import Spherical, get_intersection_locations
from upstage_des.motion.cartesian_model import cartesian_linear_intersection as cli
from upstage_des.motion.geodetic_model import analytical_intersection as agi
from upstage_des.motion.geodetic_model import subdivide_intersection as gi
from upstage_des.type_help import TASK_GEN

LOC = TypeVar("LOC", bound=UP.CartesianLocation | UP.GeodeticLocation)


def close(a: float, b: float) -> bool:
    """Test if two numbers are close.

    Args:
        a (float): number
        b (float): other number

    Returns:
        bool: if they are close
    """
    d = abs(a - b)
    return d <= 1e-8


class DummySensor(Generic[LOC]):
    """A simple sensor for testing purposes."""

    def __init__(self, env: SIM.Environment, location: LOC, radius: float = 1.0) -> None:
        self.env = env
        self.data: list[tuple[Any, float, str]] = []
        self._location = location
        self._radius = radius

    def entity_entered_range(self, mover: Any) -> None:
        self.data.append((mover, self.env.now, "detect"))
        if hasattr(mover, "loc"):
            # call the location to record it
            mover.loc

    def entity_exited_range(self, mover: Any) -> None:
        self.data.append((mover, self.env.now, "end detect"))
        if hasattr(mover, "loc"):
            # call the location to record it
            mover.loc

    @property
    def location(self) -> LOC:
        return self._location

    @property
    def radius(self) -> float:
        return self._radius


class BadSensor:
    """An incomplete sensor for testing purposes."""

    def __init__(self, env: SIM.Environment, location: tuple[float, ...], radius: float) -> None:
        self.env = env
        self._location = UP.CartesianLocation(*location)
        self._radius = radius

    @property
    def location(self) -> UP.CartesianLocation:
        return self._location

    @property
    def radius(self) -> float:
        return self._radius


class DummyMover:
    """A simple mover for testing purposes."""

    def __init__(self, env: SIM.Environment) -> None:
        self.env = env
        self.detect = True

    def _get_detection_state(self) -> str:
        return "detect"


class RealMover(UP.Actor):
    """A more realistic mover that moves in Cartesian Space."""

    loc = UP.CartesianLocationChangingState(recording=True)
    speed = UP.State[float]()
    detect = UP.DetectabilityState()


class RealGeodeticMover(UP.Actor):
    """A more realistic mover that moves in Geodetic Space."""

    loc = UP.GeodeticLocationChangingState(recording=True)
    speed = UP.State[float]()
    detect = UP.DetectabilityState()


# This is not best practices, but it works for testing
class DoMove(UP.Task):
    """A task for movers to move."""

    waypoints: list[UP.GeodeticLocation] | list[UP.CartesianLocation] = []

    def task(self, *, actor: RealGeodeticMover | RealMover) -> TASK_GEN:
        dist = 0.0
        wayps = [actor.loc] + list(self.waypoints)
        for i in range(len(wayps) - 1):
            dist += wayps[i + 1] - wayps[i]
        time = dist / actor.speed

        actor.activate_location_state(
            state="loc",
            task=self,
            speed=actor.speed,
            waypoints=self.waypoints,
        )
        yield UP.Wait(time)
        actor.deactivate_all_states(task=self)

    def on_interrupt(
        self, *, actor: RealGeodeticMover | RealMover, cause: Any
    ) -> UP.InterruptStates:
        if cause == "Become undetectable":
            actor.detect = False
            return self.INTERRUPT.IGNORE
        return self.INTERRUPT.END


L = TypeVar("L")


def _create_mover_and_waypoints(
    env: SIM.Environment,
    mover_type: type,
    location_type: type[L],
    *waypoints: tuple[float, ...],
) -> tuple[UP.Actor, list[L]]:
    mover = cast(UP.Actor, mover_type(env))
    waypoints = [location_type(*waypoint) for waypoint in waypoints]
    return mover, waypoints


def test_errors() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli)

        mover, waypoints = _create_mover_and_waypoints(
            env,
            DummyMover,
            UP.CartesianLocation,
            (0, 0, 0),
            (1, 1, 0),
        )

        bad_sensor = BadSensor(env, (0.9, 0.9), 0.5)

        with pytest.raises(UP.MotionAndDetectionError):
            motion._stop_mover(mover)

        motion._start_mover(mover, speed=1.0, waypoints=waypoints)
        with pytest.raises(UP.MotionAndDetectionError):
            motion._start_mover(mover, speed=1.0, waypoints=[[2, 2], [3, 3]])  # type: ignore [arg-type]

        with pytest.raises(NotImplementedError):
            motion.add_sensor(bad_sensor, "location", "radius")  # type: ignore [arg-type]


def test_no_interaction_cli() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli)
        mover, waypoints = _create_mover_and_waypoints(
            env,
            DummyMover,
            UP.CartesianLocation,
            (2, 0, 0),
            (2, -2, 0),
        )
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc)
        motion.add_sensor(
            sensor,
        )
        motion._start_mover(mover, 1.0, waypoints)
        env.run(until=5.0)

        assert abs(env.now - 5.0) < 1e-12

        assert len(sensor.data) == 0, (
            f"There should be no interaction events, but found: {sensor.data}"
        )
        assert not motion._events.get(mover, [])
        assert sensor not in motion._in_view.get(mover, {})
        assert motion._debug_log == [], "No log expected for no actions"


def test_enter_exit() -> None:
    with UP.EnvironmentContext() as env:
        motion = UP.SensorMotionManager(cli)
        mover, waypoints = _create_mover_and_waypoints(
            env,
            DummyMover,
            UP.CartesianLocation,
            (2, 2, 0),
            (-2, -2, 0),
        )
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc)
        motion.add_sensor(
            sensor,
        )
        motion._start_mover(mover, 1.0, waypoints)
        env.run()
        motion._stop_mover(mover)

        assert abs(env.now - 3.828427124746188) < 1e-12

        assert len(sensor.data) == 2, (
            f"For now, motion manager only has entry event recorded, {sensor.data}"
        )

        assert abs(sensor.data[0][1] - 1.8284271247461907) < 1e-12
        assert sensor.data[0][2] == "detect"
        assert abs(sensor.data[1][1] - 3.828427124746188) < 1e-12
        assert sensor.data[1][2] == "end detect"
        assert not motion._events.get(mover, [])
        assert sensor not in motion._in_view.get(mover, {})


def test_sensor_popup() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli)
        mover, waypoints = _create_mover_and_waypoints(
            env,
            DummyMover,
            UP.CartesianLocation,
            (2, 2, 0),
            (-2, -2, 0),
        )
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc)
        motion._start_mover(mover, 1.0, waypoints)
        env.run(until=2)

        motion.add_sensor(
            sensor,
        )
        env.run()
        motion._stop_mover(mover)

        assert abs(env.now - 3.828427124746188) < 1e-12

        assert len(sensor.data) == 2, (
            f"For now, motion manager only has entry event recorded, {sensor.data}"
        )

        assert abs(sensor.data[0][1] - 2) < 1e-12
        assert sensor.data[0][2] == "detect"
        assert abs(sensor.data[1][1] - 3.828427124746188) < 1e-12
        assert sensor.data[1][2] == "end detect"
        assert not motion._events.get(mover, [])
        assert sensor not in motion._in_view.get(mover, {})


def test_start_inside_exit() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli)
        mover, waypoints = _create_mover_and_waypoints(
            env,
            DummyMover,
            UP.CartesianLocation,
            (0.5, 0.5, 0),
            (-2, -2, 0),
        )
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc)
        motion.add_sensor(
            sensor,
        )
        motion._start_mover(mover, 1.0, waypoints)
        env.run()
        motion._stop_mover(mover)

        assert env.now == 1.7071067811865475

        assert len(sensor.data) == 2, "Need entry and exit events"

        assert abs(sensor.data[0][1] - 0) < 1e-12
        assert sensor.data[0][2] == "detect"
        assert abs(sensor.data[1][1] - 1.7071067811865475) < 1e-12
        assert sensor.data[1][2] == "end detect"
        assert not motion._events.get(mover, [])
        assert sensor not in motion._in_view.get(mover, {})


def test_enter_end_inside_then_leave() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli)
        mover, waypoints = _create_mover_and_waypoints(
            env,
            DummyMover,
            UP.CartesianLocation,
            (2, 2, 0),
            (-0.5, -0.5, 0),
        )
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc)
        motion.add_sensor(
            sensor,
        )
        motion._start_mover(mover, 1.0, waypoints)
        env.run()
        motion._stop_mover(mover)

        assert env.now == 1.8284271247461907

        assert len(sensor.data) == 1, "Ending inside means no exit event"

        assert abs(sensor.data[-1][1] - 1.8284271247461907) < 1e-12
        assert sensor in motion._in_view[mover]
        assert not motion._events.get(mover, [])

        # have the mover leave and run the clock a bit in case
        env.run(until=20)
        assert len(sensor.data) == 1, "No new events should be added to the log"
        assert abs(sensor.data[-1][1] - 1.8284271247461907) < 1e-12
        assert sensor in motion._in_view[mover]

        motion._start_mover(mover, 1.0, waypoints[::-1])
        env.run()
        assert env.now == 21.707106781186546
        assert len(sensor.data) == 2, "Wrong amount of sensor data"
        assert abs(sensor.data[-1][1] - 21.707106781186546) < 1e-12
        assert sensor not in motion._in_view[mover]


def test_start_inside_end_inside() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli)
        mover, waypoints = _create_mover_and_waypoints(
            env,
            DummyMover,
            UP.CartesianLocation,
            (0.5, 0.5, 0),
            (-0.5, -0.5, 0),
        )
        mover = cast(UP.Actor, mover)
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc)
        motion.add_sensor(
            sensor,
        )
        motion._start_mover(mover, 1.0, waypoints)
        env.run()

        assert env.now == 0

        assert len(sensor.data) == 1, f"Only need to see the start recorded: {sensor.data}"
        assert sensor.data[0][2] == "detect"
        assert sensor in motion._in_view[mover]


def test_motion_setup_cli() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli, debug=True)
        UP.add_stage_variable("motion_manager", motion)
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc, radius=10.0)

        mover = cast(UP.Actor, DummyMover(env))
        mover_start = UP.CartesianLocation(*[8, 8, 2])
        waypoints = [
            mover_start,
            UP.CartesianLocation(*[-8, 8, 2]),
            UP.CartesianLocation(*[-8, -8, 2]),
            UP.CartesianLocation(*[8, -8, 0]),
        ]

        motion.add_sensor(sensor, "location", "radius")
        # This isn't how these things are normally called for movers, but
        # since it isn't going through a task, it's ok here.
        motion._start_mover(mover, 1.0, waypoints)
        assert len(motion._events) == 1
        assert len(motion._events[mover]) == 3
        env.run()
        assert len(sensor.data) == 6
        matches = [2.343145750507622, 18.343145750507624, 34.268912605325006]
        for datum, truth in zip(sensor.data[::2], matches):
            assert pytest.approx(truth) == datum[1]


def test_late_intersection() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli, debug=True)
        UP.add_stage_variable("motion_manager", motion)
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc, radius=5.0)

        mover = cast(UP.Actor, DummyMover(env))
        mover_start = UP.CartesianLocation(*[0, 8, 0])
        waypoints = [
            mover_start,
            UP.CartesianLocation(*[0, 6, 0]),
        ]
        motion.add_sensor(sensor)
        # This isn't how these things are normally called for movers, but
        # since it isn't going through a task, it's ok here.
        motion._start_mover(mover, 1.0, waypoints)
        env.run()


def test_motion_coordination_cli() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli, debug=True)
        UP.add_stage_variable("motion_manager", motion)
        # Test that if the location is a changing state, that it matches up
        # when detections happen
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc, radius=10.0)

        mover_start = UP.CartesianLocation(*[8, 8, 2])
        mover = RealMover(name="A Mover", loc=mover_start, speed=1, detect=True)
        waypoints = [
            mover_start,
            UP.CartesianLocation(*[-8, 8, 2]),
            UP.CartesianLocation(*[-8, -8, 2]),
            UP.CartesianLocation(*[8, -8, 0]),
        ]
        task = DoMove()
        task.waypoints = waypoints[1:]
        task.run(actor=mover)

        motion.add_sensor(sensor)

        env.run()
        assert mover.loc == waypoints[-1]
        assert len(motion._debug_data[mover]) == 3
        for i, data in enumerate(motion._debug_data[mover]):
            sense, kinds, times, inters = data
            mover_at_time_0 = [x[1] for x in mover._state_histories["loc"] if close(x[0], times[0])]
            mover_at_time_1 = [x[1] for x in mover._state_histories["loc"] if close(x[0], times[1])]
            assert mover_at_time_0
            assert mover_at_time_1
            assert close(mover_at_time_0[0], inters[0])
            assert close(mover_at_time_1[0], inters[1])


def test_background_motion() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli, debug=True)
        UP.add_stage_variable("motion_manager", motion)
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc, radius=10.0)

        mover_start = UP.CartesianLocation(*[8, 8, 2])
        mover = RealMover(name="A Mover", loc=mover_start, speed=1, detect=True)
        waypoints = [
            mover_start,
            UP.CartesianLocation(*[-8, 8, 2]),
            UP.CartesianLocation(*[-8, -8, 2]),
            UP.CartesianLocation(*[8, -8, 0]),
        ]
        task = DoMove()
        task.waypoints = waypoints[1:]
        task.run(actor=mover)

        motion.add_sensor(sensor)
        # no need to start the mover this time
        env.run()
        assert mover.loc == waypoints[-1]
        # The motion manager needs to see the mover 3 times, and
        # so does the sensor
        assert len(motion._debug_data[mover]) == 3
        assert len(sensor.data) == 6


def test_background_rehearse() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli, debug=True)
        UP.add_stage_variable("motion_manager", motion)

        # This is the key change relative to the above
        flyer_start = UP.CartesianLocation(*[8, 8, 2])
        flyer = RealMover(name="A Mover", loc=flyer_start, speed=1, detect=True)
        waypoints = [
            flyer_start,
            UP.CartesianLocation(*[-8, 8, 2]),
            UP.CartesianLocation(*[-8, -8, 2]),
            UP.CartesianLocation(*[8, -8, 0]),
        ]
        task = DoMove()
        task.waypoints = waypoints[1:]

        flyer_clone = task.rehearse(actor=flyer)

        env.run()
        assert env.now == 0, "No time should pass for rehearsal"
        assert flyer.loc == waypoints[0]
        # The motion manager shouldn't see anything
        assert flyer not in motion._debug_data
        assert flyer_clone not in motion._debug_data


def test_interrupt_clean() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli, debug=True)
        UP.add_stage_variable("motion_manager", motion)
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc, radius=10.0)

        mover_start = UP.CartesianLocation(*[8, 8, 2])
        mover = RealMover(name="A Mover", loc=mover_start, speed=1, detect=True, debug_log=True)
        waypoints = [
            mover_start,
            UP.CartesianLocation(*[-8, 8, 2]),
            UP.CartesianLocation(*[-8, -8, 2]),
            UP.CartesianLocation(*[8, -8, 0]),
        ]
        task = DoMove()
        task.waypoints = waypoints[1:]
        proc = task.run(actor=mover)

        motion.add_sensor(sensor)
        # no need to start the mover this time
        env.run(until=25)
        proc.interrupt(cause="Stop")
        env.run()

        # The motion manager needs to see the mover 3 times, and
        # the sensor will see it double that for entry/exit
        assert len(motion._debug_data[mover]) == 3
        assert len(sensor.data) == 3
        # the two messages for cancelling the notifications
        assert len(motion._debug_log) == 5
        assert all(
            log_entry["event"] == "Scheduling sensor detecting mover"
            for log_entry in motion._debug_log[:3]
        )
        assert all(line["mover"] is mover for line in motion._debug_log)
        # the order of these two may switch..
        msgs = [
            "Detection of a mover cancelled before exit",
            "Detection of a mover cancelled before entry",
        ]
        events = [motion._debug_log[i]["event"] for i in [3, 4]]
        assert len(set(events)) == 2, "Need two unique event descriptions"
        assert all(x in msgs for x in events), "Need both types exactly"
        # the mover should be stopped
        assert mover not in motion._movers


def test_undetectable_cli() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli, debug=True)
        UP.add_stage_variable("motion_manager", motion)
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc, radius=10.0)

        mover_start = UP.CartesianLocation(*[8, 8, 2])
        mover = RealMover(name="A Mover", loc=mover_start, speed=1, debug_log=True, detect=True)
        waypoints = [
            mover_start,
            UP.CartesianLocation(*[-8, 8, 2]),
            UP.CartesianLocation(*[-8, -8, 2]),
            UP.CartesianLocation(*[8, -8, 0]),
        ]
        task = DoMove()
        task.waypoints = waypoints[1:]
        proc = task.run(actor=mover)

        motion.add_sensor(sensor)
        # no need to start the mover this time
        env.run(until=25)

        # check that the motion manager has the right data
        assert mover in motion._in_view, "Mover not found in progress"
        proc.interrupt(cause="Become undetectable")
        env.run()

        assert sensor.data[-1] == (mover, 25, "end detect")


def test_redetectable() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli, debug=True)
        UP.add_stage_variable("motion_manager", motion)
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc, radius=10.0)

        mover_start = UP.CartesianLocation(*[8, 8, 2])
        mover = RealMover(name="A Mover", loc=mover_start, speed=1, debug_log=True, detect=False)
        waypoints = [
            mover_start,
            UP.CartesianLocation(*[-8, 8, 2]),
            UP.CartesianLocation(*[-8, -8, 2]),
            UP.CartesianLocation(*[8, -8, 0]),
        ]
        task = DoMove()
        task.waypoints = waypoints[1:]
        task.run(actor=mover)

        motion.add_sensor(sensor)
        # no need to start the mover this time
        env.run(until=25)
        with pytest.warns(UserWarning, match="Setting DetectabilityState to True while*"):
            mover.detect = True


def test_undetectable_after() -> None:
    with UP.EnvironmentContext() as env:
        UP.add_stage_variable("distance_units", "m")
        motion = UP.SensorMotionManager(cli, debug=True)
        UP.add_stage_variable("motion_manager", motion)
        loc = UP.CartesianLocation(0, 0, 0)
        sensor = DummySensor(env, loc, radius=10.0)

        mover_start = UP.CartesianLocation(*[8, 8, 2])
        mover = RealMover(name="A Mover", loc=mover_start, speed=1, debug_log=True, detect=True)
        waypoints = [
            mover_start,
            UP.CartesianLocation(*[-8, 8, 2]),
            UP.CartesianLocation(*[-8, -8, 2]),
            UP.CartesianLocation(*[8, -8, 0]),
        ]
        task = DoMove()
        task.waypoints = waypoints[1:]
        proc = task.run(actor=mover)

        motion.add_sensor(sensor)
        # no need to start the mover this time
        env.run(until=30)

        # check that the motion manager has the right data
        assert mover in motion._in_view, "Mover not found in progress"
        proc.interrupt(cause="Become undetectable")
        env.run()

        assert (mover, 30, "end detect") not in sensor.data
        assert len(sensor.data) == 4


def test_motion_setup_gi() -> None:
    with UP.EnvironmentContext() as env:
        motion = UP.SensorMotionManager(gi, debug=True)
        UP.add_stage_variable("stage_model", Spherical)
        UP.add_stage_variable("altitude_units", "m")
        UP.add_stage_variable("distance_units", "nmi")
        UP.add_stage_variable("intersection_model", get_intersection_locations)
        loc = UP.GeodeticLocation(0, 0, 0)
        sensor = DummySensor(env, loc, radius=150.0)

        geo_mover = cast(UP.Actor, DummyMover(env))
        t = 2
        geo_mover_start = UP.GeodeticLocation(*[t, t, 4000])
        waypoints = [
            geo_mover_start,
            UP.GeodeticLocation(*[-t, t, 4000]),
            UP.GeodeticLocation(*[-t, -t, 4000]),
            UP.GeodeticLocation(*[t, -t, 0]),
        ]

        motion.add_sensor(sensor)
        motion._start_mover(geo_mover, 1.0, waypoints)
        assert len(motion._events) == 1
        assert len(motion._events[geo_mover]) == 3
        env.run()
        assert len(sensor.data) == 6
        matches = [30.59361210120766, 271.030439713508, 511.28446115022086]
        for datum, truth in zip(sensor.data[::2], matches):
            assert pytest.approx(truth, abs=0.001) == datum[1]


def test_no_interaction_gi() -> None:
    with UP.EnvironmentContext() as env:
        motion = UP.SensorMotionManager(gi, debug=True)
        UP.add_stage_variable("stage_model", Spherical)
        UP.add_stage_variable("altitude_units", "ft")
        UP.add_stage_variable("distance_units", "nmi")
        UP.add_stage_variable("intersection_model", get_intersection_locations)

        motion = UP.SensorMotionManager(gi)
        loc = UP.GeodeticLocation(*[90, 40, 0])
        sensor = DummySensor(env, loc, 1.0)

        mover = cast(UP.Actor, DummyMover(env))
        waypoints = [
            UP.GeodeticLocation(0, 10, 0),
            UP.GeodeticLocation(10, 10, 0),
        ]

        motion.add_sensor(
            sensor,
        )
        motion._start_mover(mover, 1.0, waypoints)
        env.run(until=5.0)

        assert abs(env.now - 5.0) < 1e-12

        assert len(sensor.data) == 0, (
            f"There should be no interaction events, but found: {sensor.data}"
        )

        motion._stop_mover(mover)


def test_motion_coordination_gi() -> None:
    with UP.EnvironmentContext() as env:
        motion = UP.SensorMotionManager(gi, debug=True)
        UP.add_stage_variable("stage_model", Spherical)
        UP.add_stage_variable("altitude_units", "m")
        UP.add_stage_variable("distance_units", "nmi")
        UP.add_stage_variable("intersection_model", get_intersection_locations)
        UP.add_stage_variable("motion_manager", motion)
        loc = UP.GeodeticLocation(0, 0, 0)
        sensor = DummySensor(env, loc, radius=150.0)

        t = 2
        geo_mover_start = UP.GeodeticLocation(*[t, t, 4000])
        geo_mover = RealGeodeticMover(name="Mover", loc=geo_mover_start, speed=1, detect=True)

        waypoints = [
            geo_mover_start,
            UP.GeodeticLocation(*[-t, t, 4000]),
            UP.GeodeticLocation(*[-t, -t, 4000]),
            UP.GeodeticLocation(*[t, -t, 0]),
        ]

        task = DoMove()
        task.waypoints = waypoints[1:]
        task.run(actor=geo_mover)

        motion.add_sensor(sensor)

        env.run()
        assert abs(geo_mover.loc - waypoints[-1]) <= 1e-12
        assert len(motion._debug_data[geo_mover]) == 3
        for i, data in zip([1, 3, 5], motion._debug_data[geo_mover]):
            sense, kinds, times, inters = data
            mover_at_time_0 = [
                x[1] for x in geo_mover._state_histories["loc"] if close(x[0], times[0])
            ]
            mover_at_time_1 = [
                x[1] for x in geo_mover._state_histories["loc"] if close(x[0], times[1])
            ]
            assert mover_at_time_0
            assert mover_at_time_1
            assert close(mover_at_time_0[0], inters[0])
            assert close(mover_at_time_1[0], inters[1])


def test_motion_setup_agi() -> None:
    with UP.EnvironmentContext() as env:
        motion = UP.SensorMotionManager(agi, debug=True)
        UP.add_stage_variable("stage_model", Spherical)
        UP.add_stage_variable("altitude_units", "m")
        UP.add_stage_variable("distance_units", "nmi")
        UP.add_stage_variable("intersection_model", get_intersection_locations)
        UP.add_stage_variable("motion_manager", motion)
        sensor = DummySensor(env, UP.GeodeticLocation(0, 0, 0), radius=150.0)

        geo_mover = cast(UP.Actor, DummyMover(env))
        t = 2
        geo_mover_start = UP.GeodeticLocation(*[t, t, 4000])
        waypoints = [
            geo_mover_start,
            UP.GeodeticLocation(*[-t, t, 4000]),
            UP.GeodeticLocation(*[-t, -t, 4000]),
            UP.GeodeticLocation(*[t, -t, 0]),
        ]

        motion.add_sensor(sensor)
        motion._start_mover(geo_mover, 1.0, waypoints)
        assert len(motion._events) == 1
        assert len(motion._events[geo_mover]) == 3
        env.run()
        assert len(sensor.data) == 6
        matches = [30.511, 270.967, 511.207]
        for datum, truth in zip(sensor.data[::2], matches):
            assert pytest.approx(truth, abs=0.1) == datum[1]


def test_no_interaction_agi() -> None:
    with UP.EnvironmentContext() as env:
        motion = UP.SensorMotionManager(agi, debug=True)
        UP.add_stage_variable("stage_model", Spherical)
        UP.add_stage_variable("altitude_units", "m")
        UP.add_stage_variable("distance_units", "nmi")
        UP.add_stage_variable("intersection_model", get_intersection_locations)
        UP.add_stage_variable("motion_manager", motion)

        motion = UP.SensorMotionManager(agi)
        sensor = DummySensor(env, UP.GeodeticLocation(0, 0, 0))
        UP.GeodeticLocation(*[0, 0, 0])

        mover = cast(UP.Actor, DummyMover(env))
        waypoints = [
            UP.GeodeticLocation(0, 10, 0),
            UP.GeodeticLocation(10, 10, 0),
        ]

        motion.add_sensor(
            sensor,
        )
        motion._start_mover(mover, 1.0, waypoints)
        env.run(until=5.0)

        assert abs(env.now - 5.0) < 1e-12

        assert len(sensor.data) == 0, (
            f"There should be no interaction events, but found: {sensor.data}"
        )

        motion._stop_mover(mover)


def test_motion_coordination_agi() -> None:
    with UP.EnvironmentContext() as env:
        motion = UP.SensorMotionManager(agi, debug=True)
        UP.add_stage_variable("stage_model", Spherical)
        UP.add_stage_variable("altitude_units", "m")
        UP.add_stage_variable("distance_units", "nmi")
        UP.add_stage_variable("intersection_model", get_intersection_locations)
        UP.add_stage_variable("motion_manager", motion)

        sensor = DummySensor(env, UP.GeodeticLocation(0, 0, 0), radius=150.0)

        t = 2
        geo_mover_start = UP.GeodeticLocation(*[t, t, 4000])
        geo_mover = RealGeodeticMover(name="Mover", loc=geo_mover_start, speed=1, detect=True)

        waypoints = [
            geo_mover_start,
            UP.GeodeticLocation(*[-t, t, 4000]),
            UP.GeodeticLocation(*[-t, -t, 4000]),
            UP.GeodeticLocation(*[t, -t, 0]),
        ]

        task = DoMove()
        task.waypoints = waypoints[1:]
        task.run(actor=geo_mover)

        motion.add_sensor(sensor)

        env.run()
        assert abs(geo_mover.loc - waypoints[-1]) <= 1e-12
        assert len(motion._debug_data[geo_mover]) == 3
        for i, data in zip([1, 3, 5], motion._debug_data[geo_mover]):
            sense, kinds, times, inters = data
            mover_at_time_0 = [
                x[1] for x in geo_mover._state_histories["loc"] if close(x[0], times[0])
            ]
            mover_at_time_1 = [
                x[1] for x in geo_mover._state_histories["loc"] if close(x[0], times[1])
            ]
            assert mover_at_time_0
            assert mover_at_time_1
            assert abs(mover_at_time_0[0] - inters[0]) <= 0.5  # nmi
            assert abs(mover_at_time_1[0] - inters[1]) <= 0.7  # nmi


def test_analytical_intersection() -> None:
    with UP.EnvironmentContext():
        UP.add_stage_variable("stage_model", Spherical)
        UP.add_stage_variable("altitude_units", "m")
        UP.add_stage_variable("distance_units", "nmi")

        start = UP.GeodeticLocation(33.67009544379275, -84.59178543542892, 5_000)
        finish = UP.GeodeticLocation(33.871012616336344, -84.16331866903882, 5_000)
        middle = UP.GeodeticLocation(33.7774620987044, -84.38304521590554, 4_000)

        res = agi(start, finish, 200.0, middle, 200.0)
        intersections, times, types, path_time = res
        assert types == ["START_INSIDE", "END_INSIDE"]



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import pytest

from upstage_des.geography import WGS84, Spherical, get_intersection_locations
from upstage_des.motion.cartesian_model import ray_intersection

from .conftest import POS


@pytest.mark.parametrize("earth", [WGS84, Spherical])
def test_intersections(
    intersect_positions: tuple[tuple[POS, POS, POS], float, str, list[str]],
    earth: WGS84 | Spherical,
) -> None:
    pos, sensor_range, range_units, answer = intersect_positions
    start_lla, finish_lla, sensor_lla = pos
    sensor_loc = (sensor_lla[0], sensor_lla[1])
    intersects = get_intersection_locations(
        start_lla,
        finish_lla,
        sensor_lla,
        sensor_range,
        range_units,
        earth,
        dist_between=9260,
        subdivide_levels=[20, 20],
    )
    # without visibility, the intersections should be very close to the right
    # range, unless they are inside the range due to START_IN
    for i, an in zip(intersects, answer):
        assert i.kind == an
        loc = (i.begin[0], i.begin[1])
        dist = earth.distance(loc, sensor_loc, range_units)
        if an in ["EXIT", "ENTER"]:
            assert sensor_range == pytest.approx(dist, rel=0.001)
        else:
            assert dist < sensor_range


@pytest.mark.parametrize("earth", [WGS84, Spherical])
def test_short_intersections(
    earth: WGS84 | Spherical, short_intersections: tuple[tuple[POS, POS, POS], float, str]
) -> None:
    pos, sensor_range, range_units = short_intersections
    start_lla, finish_lla, sensor_lla = pos
    _ = get_intersection_locations(
        start_lla,
        finish_lla,
        sensor_lla,
        sensor_range,
        range_units,
        earth,
        dist_between=9260,
        subdivide_levels=[20, 20],
    )


def test_ray_trace() -> None:
    input_1 = ((0, 2), (0, 1.8), (0, 0), (1, 1))
    input_2 = ((0, 2), (1, 2), (0, 0), (1, 1))
    input_3 = ((0, 2, 0), (0, 1.8, 0), (0, 0, 0), (1, 1, 1))
    input_4 = ((0, 2, 0), (1, 2, 0), (0, 0, 0), (1, 1, 1))

    points1, _ = ray_intersection(*input_1, 1.0)
    for x, y in points1:
        assert pytest.approx(0) == x
        assert pytest.approx(1) == abs(y)

    points2, _ = ray_intersection(*input_2, 1.0)
    assert not points2

    points3, _ = ray_intersection(*input_3, 1.0)
    for x, y, z in points3:
        assert pytest.approx(0) == x
        assert pytest.approx(1) == abs(y)

    points4, _ = ray_intersection(*input_4, 1.0)
    assert not points4



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""0A 3-D intersection model for cartesian motion in UPSTAGE along straight-line paths."""

from typing import TypeVar

from upstage_des.base import MotionAndDetectionError
from upstage_des.data_types import CartesianLocation
from upstage_des.math_utils import (
    _col_mat_mul,
    _roots,
    _vector_add,
    _vector_dot,
    _vector_norm,
    _vector_subtract,
)

XY = tuple[float, float]
XYZ = tuple[float, float, float]

L = TypeVar("L", XY, XYZ)


def ray_intersection(
    start: L,
    toward: L,
    center: L,
    radii: float | L,
    speed: float,
) -> tuple[list[L], list[float]]:
    """Ray intersection with ellispoid.

    Args:
        start (XY | XYZ): Start position
        toward (XY | XYZ): Point being looked at
        center (XY | XYZ): Center of the ellipsoid
        radii (float | XY | XYZ): Radii of each dimension of ellipsoid
        speed (float): Speed of the particle moving from start to toward

    Returns:
        tuple[list[XY] | list[XYZ], list[float]]: Intersecting positions and times.
    """
    n_dim = len(start)
    for compare in [toward, center]:
        assert len(compare) == n_dim, "Mismatched dimensions in ray intersection."

    M = [[0.0 for _ in range(n_dim)] for _ in range(n_dim)]
    _radii: tuple[float, ...]
    if isinstance(radii, float | int):
        _radii = tuple([float(radii)] * n_dim)
    elif len(radii) == 1:
        _radii = tuple([radii[0]] * n_dim)
    else:
        _radii = radii
        if len(start) != len(radii):
            raise MotionAndDetectionError(
                "Radius for a sensor must be a single float or the same dimensionality as the "
                "locations."
            )
    for i in range(n_dim):
        M[i][i] = 1 / _radii[i]

    v = _vector_subtract(toward, start)
    v1 = _col_mat_mul(v, M)
    vec1 = _col_mat_mul(start, M)
    vec2 = _col_mat_mul(center, M)
    P1 = _vector_subtract(vec1, vec2)
    c = _vector_norm(P1) ** 2 - 1
    b = 2 * _vector_dot(P1, v1)
    a = _vector_norm(v1) ** 2
    roots = _roots(a, b, c)
    possible = [r for r in roots if r >= 0]

    intersections: list[L] = []
    times: list[float] = []
    for t in sorted(possible):
        loc = _vector_add(start, [t * x for x in v])
        dist = float(_vector_norm(_vector_subtract(loc, start)))
        times.append(dist / speed)
        intersections.append(tuple(loc))  # type: ignore [arg-type]
    return intersections, times


def cartesian_linear_intersection(
    start: CartesianLocation,
    finish: CartesianLocation,
    speed: float,
    sensor_location: CartesianLocation,
    sensor_radius: float,
) -> tuple[list[CartesianLocation], list[float], list[str], float]:
    """Get the intersection of straight line motion and a sphere.

    Args:
        start (CartesianLocation): the start location of the mover
        finish (CartesianLocation): the finish location of the mover
        speed (float): the speed of the mover (in STAGE units)
        sensor_location (CartesianLocation): the location of the sensor
        sensor_radius (float): the radius of the sensor (in STAGE units)

    Returns:
        tuple[list[CartesianLocation], list[float], list[str], float]: _description_
    """
    path_dist = finish - start
    path_time = path_dist / speed
    dist_start = sensor_location - start
    dist_finish = sensor_location - finish
    start_tup = (start.x, start.y, start.z)
    finish_tup = (finish.x, finish.y, finish.z)

    # for start and finish inside, we can skip the intersection math
    start_inside = dist_start <= sensor_radius
    finish_inside = dist_finish <= sensor_radius
    if start_inside and finish_inside:
        inters = [start, finish]
        the_times = [0, path_time]
        return inters, the_times, ["START_INSIDE", "END_INSIDE"], path_time

    sensor_location_tup = (sensor_location.x, sensor_location.y, sensor_location.z)
    intersections, times = ray_intersection(
        start_tup,
        finish_tup,
        sensor_location_tup,
        sensor_radius,
        speed,
    )
    if not intersections or min(times) > path_time:
        return [], [], ["BAD", "BAD"], -1.0

    # filter out intersections beyond the path length
    idxs = sorted(range(len(times)), key=lambda i: times[i])
    inters = [CartesianLocation(*intersections[i]) for i in idxs if times[i] <= path_time]
    times = [times[i] for i in idxs if times[i] <= path_time]
    type_start = "START_INSIDE" if start_inside else "ENTER"
    type_end = "END_INSIDE" if finish_inside else "EXIT"
    if start_inside:
        inters = [start] + inters
        times = [0] + times
    if finish_inside:
        inters = inters + [finish]
        times = times + [path_time]
    if len(inters) != 2:
        raise MotionAndDetectionError(
            f"There should be two intersections, not {len(inters)} ({inters})"
        )
    if len(times) != 2:
        raise MotionAndDetectionError("There should be two time intersections")

    return inters, times, [type_start, type_end], path_time



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Fixtures for testing."""

from typing import Any

import pytest

import upstage_des.api as UP


@pytest.fixture
def base_actors() -> tuple[tuple[UP.State, ...], tuple[type[UP.Actor], ...]]:
    """State and Actor classes for testing.

    Returns:
        tuple[tuple[UP.State, ...], tuple[UP.Actor, ...]]: States and Actors.
    """
    first_state = UP.State[Any]()
    second_state = UP.State[Any]()
    third_state = UP.State[Any]()
    fourth_state = UP.State[Any]()

    class ActorSubclass(UP.Actor):
        state_one = first_state
        state_two = second_state

        def a_function(self, inp: Any) -> tuple[UP.Actor, Any]:
            return self, inp

    class DoubleSubclass(ActorSubclass):
        state_three = third_state
        state_four = fourth_state

        def b_function(self, inp: Any) -> tuple[UP.Actor, Any]:
            return self, inp

    states = (first_state, second_state, third_state, fourth_state)
    actors = (ActorSubclass, DoubleSubclass)

    return states, actors


@pytest.fixture
def task_objects() -> tuple[type[UP.TerminalTask], type[UP.TerminalTask], type[UP.Actor]]:
    """Example task objects for testing.

    Returns:
        tuple[UP.Task, UP.Task, UP.Actor]: The task objects.
    """

    class EndPoint(UP.TerminalTask):
        def log_message(self, *, actor: UP.Actor) -> str:
            return "The Message"

    class EndPointBase(UP.TerminalTask):
        pass

    class Dummy(UP.Actor):
        status = UP.State[Any]()

    return EndPoint, EndPointBase, Dummy



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from typing import Any

import pytest

from upstage_des.api import (
    Actor,
    CartesianLocationChangingState,
    EnvironmentContext,
    GeodeticLocationChangingState,
    InterruptStates,
    LinearChangingState,
    State,
    Task,
    Wait,
    add_stage_variable,
)
from upstage_des.data_types import CartesianLocation, GeodeticLocation
from upstage_des.geography import Spherical
from upstage_des.type_help import TASK_GEN


class Mover(Actor):
    location = CartesianLocationChangingState(recording=True)
    speed = State[float](recording=True)
    fuel = LinearChangingState(recording=True)
    fuel_burn = State[float](recording=True)

    def get_distance(self, waypoints: list[CartesianLocation]) -> float:
        d = waypoints[0] - self.location
        for i in range(1, len(waypoints)):
            d += waypoints[i] - waypoints[i - 1]
        return d


class MoverGeo(Actor):
    location = GeodeticLocationChangingState(recording=True)
    speed = State[float](recording=True)
    fuel = LinearChangingState(recording=True)
    fuel_burn = State[float](recording=True)

    def get_distance(self, waypoints: list[GeodeticLocation]) -> float:
        d = waypoints[0] - self.location
        for i in range(1, len(waypoints)):
            d += waypoints[i] - waypoints[i - 1]
        return d


class MoveTask(Task):
    def task(self, *, actor: Mover | MoverGeo) -> TASK_GEN:
        destinations = list(self.get_actor_knowledge(actor, "destinations"))
        actor.activate_state(
            state="location",
            task=self,
            speed=actor.speed,
            waypoints=destinations,
        )
        actor.activate_state(
            state="fuel",
            task=self,
            rate=actor.fuel_burn,
        )
        dist = actor.get_distance(destinations)
        time = dist / actor.speed
        yield Wait(time)
        actor.deactivate_all_states(task=self)

    def on_interrupt(self, *, actor: Mover | MoverGeo, cause: Any) -> InterruptStates:
        if cause == "restart":
            rem_wypts = actor.get_remaining_waypoints(
                location_state="location",
            )
            self.set_actor_knowledge(
                actor,
                "destinations",
                rem_wypts,
                overwrite=True,
            )
            actor.speed /= 2
            return self.INTERRUPT.RESTART
        else:
            return self.INTERRUPT.END


WAYPOINTS = [
    (3, 4),
    (6, 0),
]

ATLANTA = [33.7490, -84.3880, 1050]
DENVER = [39.7392, -104.9903, 30_000]
SAN_FRAN = [37.7749, -122.4194, 0]

WAYPOINTS_GEO = [
    DENVER,
    SAN_FRAN,
]


def test_regular_run() -> None:
    with EnvironmentContext() as env:
        add_stage_variable("distance_units", "nmi")
        add_stage_variable("altitude_units", "ft")
        add_stage_variable("stage_model", Spherical)

        mover = Mover(
            name="example",
            location=CartesianLocation(0, 0),
            speed=10,
            fuel=500,
            fuel_burn=10,
        )
        waypoints = [CartesianLocation(*x) for x in WAYPOINTS]
        mover.set_knowledge("destinations", waypoints)
        task = MoveTask()
        task.run(actor=mover)
        env.run()
        assert env.now == 1
        d = mover.location - waypoints[-1]
        assert pytest.approx(d) == 0


def test_first_interrupt() -> None:
    with EnvironmentContext() as env:
        add_stage_variable("altitude_units", "ft")
        add_stage_variable("distance_units", "nmi")
        add_stage_variable("stage_model", Spherical)

        mover = Mover(
            name="example",
            location=CartesianLocation(0, 0),
            speed=10,
            fuel=500,
            fuel_burn=10,
        )
        waypoints = [CartesianLocation(*x) for x in WAYPOINTS]
        mover.set_knowledge("destinations", waypoints)
        task = MoveTask()
        proc = task.run(actor=mover)
        env.run(until=0.25)
        proc.interrupt(cause="restart")
        env.run()
        assert env.now == (2 - 0.25)
        d = mover.location - waypoints[-1]
        assert pytest.approx(d) == 0


def test_second_interrupt() -> None:
    with EnvironmentContext() as env:
        add_stage_variable("altitude_units", "ft")
        add_stage_variable("distance_units", "nmi")
        add_stage_variable("stage_model", Spherical)

        mover = Mover(
            name="example",
            location=CartesianLocation(0, 0),
            speed=10,
            fuel=500,
            fuel_burn=10,
        )
        waypoints = [CartesianLocation(*x) for x in WAYPOINTS]
        mover.set_knowledge("destinations", waypoints)
        task = MoveTask()
        proc = task.run(actor=mover)
        env.run(until=0.75)
        proc.interrupt(cause="restart")
        env.run()
        assert env.now == (2 - 0.75)
        d = mover.location - waypoints[-1]
        assert pytest.approx(d) == 0


def test_regular_run_geo() -> None:
    with EnvironmentContext() as env:
        add_stage_variable("altitude_units", "ft")
        add_stage_variable("distance_units", "nmi")
        add_stage_variable("stage_model", Spherical)

        mover = MoverGeo(
            name="example",
            location=GeodeticLocation(*SAN_FRAN),
            speed=10,
            fuel=500000,
            fuel_burn=10,
        )
        waypoints = [GeodeticLocation(*x) for x in WAYPOINTS_GEO]
        mover.set_knowledge("destinations", waypoints)
        task = MoveTask()
        task.run(actor=mover)
        env.run()
        assert env.now == 164.81767330428073
        d = mover.location - waypoints[-1]
        assert pytest.approx(d) == 0


def test_first_interrupt_geo() -> None:
    with EnvironmentContext() as env:
        add_stage_variable("altitude_units", "ft")
        add_stage_variable("distance_units", "nmi")
        add_stage_variable("stage_model", Spherical)

        mover = MoverGeo(
            name="example",
            location=GeodeticLocation(*SAN_FRAN),
            speed=10,
            fuel=500000,
            fuel_burn=10,
        )
        waypoints = [GeodeticLocation(*x) for x in WAYPOINTS_GEO]
        mover.set_knowledge("destinations", waypoints)
        task = MoveTask()
        proc = task.run(actor=mover)
        env.run(until=40)
        proc.interrupt(cause="restart")
        env.run()
        assert pytest.approx(env.now) == (164.81767330428073 * 2 - 40)
        d = mover.location - waypoints[-1]
        assert pytest.approx(d) == 0


def test_second_interrupt_geo() -> None:
    with EnvironmentContext() as env:
        add_stage_variable("altitude_units", "ft")
        add_stage_variable("distance_units", "nmi")
        add_stage_variable("stage_model", Spherical)

        mover = MoverGeo(
            name="example",
            location=GeodeticLocation(*SAN_FRAN),
            speed=10,
            fuel=500000,
            fuel_burn=10,
        )
        waypoints = [GeodeticLocation(*x) for x in WAYPOINTS_GEO]
        mover.set_knowledge("destinations", waypoints)
        task = MoveTask()
        proc = task.run(actor=mover)
        env.run(until=130)
        proc.interrupt(cause="restart")
        env.run()
        assert pytest.approx(env.now) == (130 + 69.635346608)
        d = mover.location - waypoints[-1]
        assert pytest.approx(d) == 0



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""This file contains a ContinuousContainer."""

from collections.abc import Callable, Generator
from typing import TYPE_CHECKING, Any

from simpy import Environment, Event, Interrupt, Process
from simpy.core import BoundClass

__all__ = (
    "ContinuousContainer",
    "ContainerEmptyError",
    "ContainerError",
    "ContainerFullError",
)

EMPTY_STATUS: str = "empty"
FULL_STATUS: str = "full"


class ContainerError(Exception):
    """The container is in an invalid state."""

    @property
    def cause(self) -> Any:
        """Get the exception's cause.

        Returns:
            Any: The cause.
        """
        return self.args[0]


class ContainerFullError(ContainerError):
    """The container has reach or exceeded its capacity."""

    pass


class ContainerEmptyError(ContainerError):
    """The container is empty or has a negative level."""

    pass


class _ContinuousEvent(Event):
    def _run(self, runtime: float) -> Generator[Event, None, None]:
        self.container.add_user(self)
        do_remove = True
        try:
            evt = self.env.timeout(runtime)
            yield evt
        except Interrupt as interruption:
            if interruption.cause == self.stop_cause:
                # The container will handle this for us
                do_remove = False
                for callback in self.custom_callbacks:
                    callback()
            elif interruption.cause != "stop":
                raise Interrupt(interruption)
        if do_remove:
            self.container.remove_user(self)

    def __init__(
        self,
        container: "ContinuousContainer",
        rate: float,
        time: float,
        stop_cause: str | None,
        custom_callbacks: list[Callable[[], None]] | None = None,
    ) -> None:
        super().__init__(container.env)
        self.container = container
        self.rate = rate
        self.custom_callbacks = custom_callbacks or []
        self.stop_cause = stop_cause
        time = float("inf") if time is None else time
        self.process = self.env.process(self._run(time))

    def cancel(self) -> None:
        """Cancel this request.

        This method has to be called if the put request must be aborted, for
        example if a process needs to handle an exception like an
        :class:`~simpy.events.Interrupt`.

        If the put request was created in a :keyword:`with` statement, this
        method is called automatically.

        """
        if self.process.is_alive:
            self.process.interrupt("stop")


class ContinuousPut(_ContinuousEvent):
    """An event that puts *rate* per unit time into the *container*.

    Raise a :exc:`ValueError` if ``rate <= 0``.
    """

    def __init__(
        self,
        container: "ContinuousContainer",
        rate: float,
        time: float,
        custom_callbacks: list[Callable[[], None]] | None = None,
    ) -> None:
        """Create a put event that is continuous.

        Args:
            container (ContinuousContainer): Container to add to.
            rate (float): Rate to add at.
            time (float): Time to run the event.
            custom_callbacks (list[Callable[[], None]] | None, optional): Callbacks
                for completion. Defaults to None.
        """
        if rate <= 0:
            raise ValueError(
                "Rates must be greater than zero. Put means 'positive'."
            )  # pragma: no cover
        super().__init__(container, rate, time, FULL_STATUS, custom_callbacks)


class ContinuousGet(_ContinuousEvent):
    """An event that gets *amount* from the *container*."""

    def __init__(
        self,
        container: "ContinuousContainer",
        rate: float,
        time: float,
        custom_callbacks: list[Callable[[], None]] | None = None,
    ) -> None:
        """Create a get event that is continuous.

        Args:
            container (ContinuousContainer): Container to take from.
            rate (float): Rate to remote at.
            time (float): Time to run the event.
            custom_callbacks (list[Callable[[], None]] | None, optional): Callbacks
                for completion. Defaults to None.
        """
        if rate <= 0:
            raise ValueError(
                "Rates must be greater than zero. Get means 'negative'."
            )  # pragma: no cover
        super().__init__(container, -rate, time, EMPTY_STATUS, custom_callbacks)


class ContinuousContainer:
    """A container that accepts continuous gets and puts."""

    def __init__(
        self,
        env: Environment,
        capacity: int | float,
        init: int | float = 0.0,
        error_empty: bool = True,
        error_full: bool = True,
    ):
        """Create a container that allows continuous gets and puts.

        Args:
            env (Environment): SimPy Environment.
            capacity (int | float): Capacity of the container
            init (int | float, optional): Initial amount. Defaults to 0.0.
            error_empty (bool, optional): Error when it gets empty. Defaults to True.
            error_full (bool, optional): Error when it gets full. Defaults to True.
        """
        self._capacity = capacity
        if init < 0 or capacity < 0:
            raise ValueError("Initial and capacity cannot be negative.")  # pragma: no cover
        self._level = init
        self.error_empty = error_empty
        self.error_full = error_full

        self._rate: float = 0.0
        self._env: Environment = env
        self._last: float = self._env.now
        self._active_users: list[_ContinuousEvent] = []
        self._checking: Process | None = None

    if TYPE_CHECKING:

        def put(
            self,
            rate: float,
            time: float,
            custom_callbacks: list[Callable[[], None]] | None = None,
        ) -> ContinuousPut:
            """Request to put *item* into the store."""
            return ContinuousPut(self, rate, time, custom_callbacks)

        def get(
            self, rate: float, time: float, custom_callbacks: list[Callable[[], None]] | None = None
        ) -> ContinuousGet:
            """Request to get an *item* out of the store."""
            return ContinuousGet(self, rate, time, custom_callbacks)

    else:
        put = BoundClass(ContinuousPut)
        get = BoundClass(ContinuousGet)

    def time_until_level(self, level: float, rate: float = 0.0) -> float:
        """Calculate the time until the containers reaches a value.

        Args:
            level (float): The value to reach.
            rate (float, optional): Additional rate. Defaults to 0.0.

        Returns:
            float: The time to reach the level.
        """
        rate += self._rate

        if self.level == level:
            return 0.0  # pragma: no cover
        elif rate == 0:
            return float("inf")  # pragma: no cover
        time = (level - self._level) / rate
        return time if time > 0 else float("inf")

    def time_until_done(self, rate: float = 0.0) -> float:
        """Calculate the time until the container is full or empty.

        Args:
            rate (float, optional): Additional rate. Defaults to 0.0.

        Returns:
            float: Time until the container reaches a limit.
        """
        rate += self._rate
        if rate > 0:
            return (self.capacity - self.level) / rate
        elif rate < 0:
            return -self.level / rate
        else:
            return float("inf")  # pragma: no cover

    @property
    def env(self) -> Environment:
        """Get the environment of the container.

        Returns:
            Environment: The SimPy environment.
        """
        return self._env

    @property
    def rate(self) -> float:
        """Get the current net rate.

        Returns:
            float: The net rate.
        """
        return self._rate

    @property
    def capacity(self) -> float:
        """Get the capacity of the container.

        Returns:
            float: The capacity.
        """
        return self._capacity

    @property
    def _active_puts(self) -> list[ContinuousPut]:
        puts = []
        for x in self._active_users:
            if x.rate > 0:
                assert isinstance(x, ContinuousPut)
                puts.append(x)
        return puts

    @property
    def _active_gets(self) -> list[ContinuousGet]:
        gets = []
        for x in self._active_users:
            if x.rate < 0:
                assert isinstance(x, ContinuousGet)
                gets.append(x)
        return gets

    def _set_level(self) -> float:
        """Set the level of the container based on the active gets/puts.

        Returns:
            float: The current level.
        """
        now = self._env.now
        if now > self._last:
            self._level += self._rate * (now - self._last)
            self._last = now
        return self._level

    def _check_empty(self) -> tuple[list[ContinuousGet], float]:
        level = self._level
        to_rem: list[ContinuousGet] = []
        rate = 0.0
        if level > 0:
            return to_rem, rate
        for get in tuple(self._active_gets):
            get.process.interrupt(EMPTY_STATUS)
            to_rem.append(get)
            rate += -get.rate
        if level == 0 and self.error_empty:
            raise ContainerEmptyError("Container is empty!")
        elif level < 0.0:
            raise ContainerError(f"Container level is less than 0 ({level:.3f})!")
        return to_rem, rate

    def _check_full(self) -> tuple[list[ContinuousPut], float]:
        level = self._level
        to_rem: list[ContinuousPut] = []
        rate: float = 0.0
        if level < self.capacity:
            return to_rem, rate
        for put in tuple(self._active_puts):
            put.process.interrupt(FULL_STATUS)
            to_rem.append(put)
            rate += -put.rate
        if level == self.capacity and self.error_full:
            raise ContainerFullError("Container is full!")
        elif level > self.capacity:
            msg = "Container level exceeds capacity by {:.3f}!"
            raise ContainerError(msg.format(level - self._capacity))
        return to_rem, rate

    def _check(self) -> Generator[Event, None, None]:
        if not self._rate or self._rate == 0:
            return  # pragma: no coverd

        level = self.level

        check_wait = ((self._capacity if self._rate > 0 else 0.0) - level) / self._rate

        try:
            yield self._env.timeout(check_wait)
        except Interrupt as interruption:
            if interruption.cause != "updated":
                raise  # pragma: no cover
        finally:
            to_rem: list[ContinuousGet] | list[ContinuousPut] = []
            rate: float
            self._set_level()
            if self._rate < 0:
                to_rem, rate = self._check_empty()
            elif self._rate > 0:
                to_rem, rate = self._check_full()
            if to_rem:
                for req in to_rem:
                    self._active_users.remove(req)
                self._add_rate(rate, interrupt=False)

    def _add_rate(self, rate_change: float | int, interrupt: bool = True) -> None:
        """Add a new rate to the existing rate."""
        if self._checking is not None and interrupt:
            self._checking.interrupt("updated")

        self._set_level()
        self._rate += rate_change

        if self._rate == 0.0:
            self._checking = None
        else:
            self._checking = self._env.process(self._check())

    def add_user(self, user: _ContinuousEvent) -> None:
        """Add a user to the container.

        Args:
            user (_ContinuousEvent): The user event
        """
        self._active_users.append(user)
        self._add_rate(user.rate)

    def remove_user(self, user: _ContinuousEvent) -> None:
        """Remove a user of the container.

        Args:
            user (_ContinuousEvent): The user event.
        """
        to_remove_rate = -1 * user.rate
        self._active_users.remove(user)
        self._add_rate(to_remove_rate)

    def _set_new_rate(self, rate: float | int) -> None:
        """Set a new rate.

        Args:
            rate (float | int): The new total rate.
        """
        curr_rate = self.rate
        diff = rate - curr_rate
        self._add_rate(diff)

    @property
    def level(self) -> float:
        """Get the level of the container.

        Returns:
            float: The current amount remaining.
        """
        return self._level + self._rate * (self._env.now - self._last)



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""A framework for modeling and simulating complex systems of systems.

UPSTAGE (i.e., the Universal Platform for Simulating Tasks and Actors with
Graphs and Events) is built atop of SimPy, with the intent of simplifying the
development process for complex simulations.

"""

from ._version import __authors__, __version__

__all__ = ("__authors__", "__version__")



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Module for testing UPSTAGE."""



import upstage_des.api as UP
from upstage_des.type_help import TASK_GEN


class KnowEvenTask(UP.Task):
    def task(self, *, actor: UP.Actor) -> TASK_GEN:
        evt = actor.create_knowledge_event(name="EvtName")
        actor.create_knowledge_event(name="other evt")
        self.set_actor_knowledge(actor, "finished", False)
        yield evt
        self.set_actor_knowledge(actor, "finished", True, overwrite=True)

    def on_interrupt(self, *, actor: UP.Actor, cause: str) -> UP.InterruptStates:
        self.set_actor_knowledge(actor, "cause", cause)
        return UP.InterruptStates.END


def test_knowledge_event_clear() -> None:
    with UP.EnvironmentContext() as env:
        act = UP.Actor(name="Example")

        task = KnowEvenTask()
        task.run(actor=act)

        env.run()
        assert not act._knowledge["finished"]
        act.succeed_knowledge_event(name="EvtName")
        env.run()
        assert act._knowledge["finished"]
        assert "EvtName" not in act._knowledge
        assert "other evt" in act._knowledge

    with UP.EnvironmentContext() as env:
        act = UP.Actor(name="Example")

        task = KnowEvenTask()
        proc = task.run(actor=act)

        env.run()
        assert not act._knowledge["finished"]
        proc.interrupt(cause="ending")
        env.run()
        assert "EvtName" not in act._knowledge
        assert not act._knowledge["finished"]
        assert act._knowledge["cause"] == "ending"
        # Only the yielded event should be cleared.
        assert "other evt" in act._knowledge


if __name__ == "__main__":
    test_knowledge_event_clear()



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from random import uniform

import pytest

from upstage_des.base import EnvironmentContext
from upstage_des.resources.container import (
    ContainerEmptyError,
    ContainerError,
    ContinuousContainer,
)
from upstage_des.resources.monitoring import (
    SelfMonitoringContainer,
    SelfMonitoringContinuousContainer,
)
from upstage_des.type_help import SIMPY_GEN

CONTAINER_CAPACITY = 100
INITIAL_LEVEL = 50


def test_basics() -> None:
    with EnvironmentContext() as env:
        con = ContinuousContainer(
            capacity=CONTAINER_CAPACITY,
            env=env,
            init=INITIAL_LEVEL,
        )

        assert con.capacity == CONTAINER_CAPACITY
        assert con.level == INITIAL_LEVEL

        env.run(until=10)
        con._set_new_rate(-10.0)

        with pytest.raises(ContainerEmptyError) as err:
            env.run(until=30)
        assert err.value.cause == "Container is empty!"


def test_error() -> None:
    with EnvironmentContext() as env:
        con = ContinuousContainer(
            capacity=CONTAINER_CAPACITY,
            env=env,
            init=INITIAL_LEVEL,
        )
        con._set_new_rate(-20)
        with pytest.raises(ContainerError):
            env.run(until=30)

        with pytest.raises(ValueError):
            con.put(-20, 3)


def test_calculations() -> None:
    with EnvironmentContext() as env:
        con = ContinuousContainer(
            capacity=CONTAINER_CAPACITY,
            env=env,
            init=INITIAL_LEVEL,
        )
        con._set_new_rate(-20.0)

        assert con.time_until_level(0.0) == 2.5
        assert con.time_until_level(CONTAINER_CAPACITY) == float("inf")


def test_checking() -> None:
    with EnvironmentContext() as env:
        auto_started = ContinuousContainer(
            capacity=CONTAINER_CAPACITY,
            env=env,
            init=INITIAL_LEVEL,
        )
        auto_started._set_new_rate(-20.0)
        env.run(until=2)
        assert auto_started.level < INITIAL_LEVEL

        with pytest.raises(ContainerEmptyError):
            env.run(until=10)


def test_get_and_put() -> None:
    with EnvironmentContext() as env:
        tank = ContinuousContainer(env, capacity=1000.0, init=500.0)
        get_time = 10.0
        get_rate = 50.0
        tank.get(rate=get_rate, time=get_time)

        put_time = 12.0
        put_rate = 60.0
        tank.put(rate=put_rate, time=put_time)

        env.run(until=9)
        assert len(tank._active_puts) == 1
        assert len(tank._active_gets) == 1

        added = -get_rate * 9 + put_rate * 9
        assert tank.level == 500 + added

        env.run(until=11)
        assert len(tank._active_puts) == 1
        assert len(tank._active_gets) == 0

        env.run(until=13)
        added = -get_rate * get_time + put_rate * put_time
        assert tank.level == 720


def test_interrupting() -> None:
    with EnvironmentContext() as env:
        tank = ContinuousContainer(
            env,
            capacity=1000.0,
            init=500.0,
            error_empty=False,
            error_full=False,
        )

        msg = []

        def full_callback() -> None:
            msg.append("full")

        putter = tank.put(10.0, 100.0, custom_callbacks=[full_callback])
        env.run(until=51)

        assert len(msg) == 1
        assert putter not in tank._active_users

    with EnvironmentContext() as env:
        tank = ContinuousContainer(
            env,
            capacity=1000.0,
            init=500.0,
            error_empty=False,
            error_full=False,
        )

        putter = tank.put(10.0, 100.0)
        env.run(until=25)
        assert tank.level == 500 + 25 * 10
        assert putter in tank._active_users
        putter.cancel()
        env.run()
        assert tank.level == 500 + 25 * 10
        assert putter not in tank._active_users


def test_complex_behavior() -> None:
    times = []
    tanker_called = [False]

    with EnvironmentContext() as env:
        tank = SelfMonitoringContinuousContainer(
            env=env,
            capacity=1000.0,
            init=850.0,
        )
        tank._set_new_rate(-1.0)

        d: list[str] = []

        def tank_is_empty() -> None:
            d.append("empty")

        def tank_is_overflowing() -> None:
            d.append("overflowing")

        def call_refill() -> SIMPY_GEN:
            yield env.timeout(5.0)

            rate = 10.0
            until = min(0.95 * tank.time_until_done(rate=rate), 5.0)

            tank.put(rate=rate, time=until, custom_callbacks=[tank_is_overflowing])
            tanker_called[0] = False
            yield env.timeout(until)

        def draw_fuel() -> SIMPY_GEN:
            while True:
                wait = uniform(10.0, 15.0)

                yield env.timeout(wait)

                rate = uniform(1.0, 2.0)
                until = 0.5 * min(tank.time_until_done(rate * 0.99), uniform(20.0, 30.0))
                # print(f"{env.now:5.1f} - Random Draw (rate: {rate:.1f}, "
                #       f"until={until:.1f})")
                tank.get(rate=rate, time=until)
                yield env.timeout(until)
                # print("{:5.1f} - Random Draw Stopped".format(env.now, rate,
                #                                              until))

        def start_stop_getting() -> SIMPY_GEN:
            getter = tank.get(rate=1.0, time=1_000, custom_callbacks=[tank_is_empty])
            # print("{:5.1f} - Started constant getting".format(env.now))
            while True:
                wait = uniform(40.0, 60.0)
                yield env.timeout(wait)
                # print(f"{env.now:5.1f} - Stopped constant getting")

                getter.process.interrupt("stop")

                wait = uniform(40.0, 60.0)
                yield env.timeout(wait)

                getter = tank.get(rate=10.0, time=1_000, custom_callbacks=[tank_is_empty])
                # print("{:5.1f} - Restarted constant getting".format(env.now))

        def simulate() -> SIMPY_GEN:
            env.process(draw_fuel())
            env.process(start_stop_getting())
            while True:
                times.append(env.now)
                t = tank.time_until_level(0.4 * tank.capacity)
                if t == float("inf"):
                    t = tank.time_until_level(tank.capacity)
                if t == float("inf"):
                    t = tank.time_until_level(0.1 * tank.capacity)
                t = 1 if t == float("inf") else t
                yield env.timeout(t)
                if not tanker_called[0] and tank.level <= 0.5 * tank.capacity:
                    tanker_called[0] = True
                    env.process(call_refill())
                yield env.timeout(1.0)

        env.process(simulate())
        env.run(until=120)
        tank._set_level()


def test_basics_monitoring() -> None:
    with EnvironmentContext() as env:
        con = SelfMonitoringContinuousContainer(
            capacity=CONTAINER_CAPACITY,
            env=env,
            init=INITIAL_LEVEL,
        )

        assert con._capacity == CONTAINER_CAPACITY
        assert con.level == INITIAL_LEVEL

        env.run(until=10)
        con._set_new_rate(-10.0)

        with pytest.raises(ContainerEmptyError):
            env.run(until=30)

        assert len(con._quantities) == 3


def test_checking_monitoring() -> None:
    with EnvironmentContext() as env:
        auto_started = SelfMonitoringContinuousContainer(
            capacity=CONTAINER_CAPACITY,
            env=env,
            init=INITIAL_LEVEL,
        )
        auto_started._add_rate(-20.0)
        env.run(until=2)
        assert auto_started.level < INITIAL_LEVEL

        with pytest.raises(ContainerEmptyError):
            env.run(until=10)


def test_self_monitoring_container() -> None:
    """Test self-monitoring container"""
    with EnvironmentContext() as env:
        con = SelfMonitoringContainer(env=env, capacity=CONTAINER_CAPACITY, init=INITIAL_LEVEL)

        assert len(con._quantities) == 1



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import simpy as SIM

import upstage_des.api as UP
from upstage_des.api import Task
from upstage_des.type_help import SIMPY_GEN, TASK_GEN


class ParallelTest(UP.Actor):
    comms = UP.State[SIM.Store]()
    logger = UP.State[list]()
    internal = UP.State[SIM.Store]()
    working = UP.State[bool]()


class TaskOne(Task):
    def task(self, *, actor: ParallelTest) -> TASK_GEN:
        actor.working = not actor.working
        actor._lock_state(state="working", task=self)
        thing = yield UP.Get(actor.internal)
        actor.logger.append((actor.env.now, thing, actor.working))
        actor._unlock_state(state="working", task=self)


class TaskTwo(Task):
    def task(self, *, actor: ParallelTest) -> TASK_GEN:
        other = yield UP.Get(actor.comms)
        actor.logger.append((actor.env.now, "Put the message", actor.working))
        yield UP.Put(actor.internal, other)


def test_parallel_looping() -> None:
    with UP.EnvironmentContext() as env:
        net_1_classes = {"Task": TaskOne}
        net_1_links = {"Task": UP.TaskLinks(default="Task", allowed=["Task"])}
        net_2_classes = {"Task": TaskTwo}
        net_2_links = {"Task": UP.TaskLinks(default="Task", allowed=["Task"])}

        tn1 = UP.TaskNetwork("InternalGet", net_1_classes, net_1_links)
        tn2 = UP.TaskNetwork("ExternalGet", net_2_classes, net_2_links)

        def proc(env: SIM.Environment, actor: ParallelTest, thing: str) -> SIMPY_GEN:
            yield env.timeout(1.3)
            yield actor.comms.put(thing)
            yield env.timeout(2.2)
            yield actor.comms.put(thing)

        pt = ParallelTest(
            name="Parallel_Actor",
            comms=SIM.Store(env),
            internal=SIM.Store(env),
            logger=[],
            working=False,
        )

        pt.add_task_network(tn1)
        pt.add_task_network(tn2)

        pt.start_network_loop("InternalGet", init_task_name="Task")
        pt.start_network_loop("ExternalGet", init_task_name="Task")
        env.process(proc(env, pt, "the msg"))
        env.run(until=1.0)
        running = pt.get_running_tasks()
        assert len(running) == 2
        assert "InternalGet" in running
        assert running["InternalGet"].name == "Task"
        assert "ExternalGet" in running
        assert running["ExternalGet"].name == "Task"

        tqs = pt.get_all_task_queues()
        assert "InternalGet" in tqs
        assert "ExternalGet" in tqs

        env.run()

        assert len(pt.logger) == 4
        expected_log = [
            (1.3, "Put the message", True),
            (1.3, "the msg", True),
            (3.5, "Put the message", False),
            (3.5, "the msg", False),
        ]
        for el, al in zip(expected_log, pt.logger):
            assert el == al



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from typing import Any

from simpy import Store

import upstage_des.api as UP
from upstage_des.api import (
    Actor,
    CommsManager,
    EnvironmentContext,
    Get,
    Message,
    MessageContent,
    ResourceState,
    State,
    Task,
    Wait,
)
from upstage_des.communications.processes import generate_comms_wait
from upstage_des.type_help import SIMPY_GEN, TASK_GEN


class ReceiveSend(Actor):
    incoming = ResourceState[Store](default=Store)
    result = State[Any](default="None")


class ReceiveTask(Task):
    def task(self, *, actor: ReceiveSend) -> TASK_GEN:
        item = yield Get(actor.incoming)
        actor.result = item


class SendTask(Task):
    comms: CommsManager
    receiver: ReceiveSend

    def task(self, *, actor: ReceiveSend) -> TASK_GEN:
        yield Wait(1.0)
        content = MessageContent(data=dict(action="move", thought="good"))
        message = Message(actor, content, self.receiver)
        yield self.comms.make_put(message, actor, self.receiver)


def test_send_receive() -> None:
    with EnvironmentContext() as env:
        receiver = ReceiveSend(name="recv")
        sender = ReceiveSend(name="send")

        rec_task = ReceiveTask()
        sen_task = SendTask()

        comms = CommsManager(
            name="Comm",
            init_entities=[(receiver, "incoming")],
            debug_logging=True,
        )
        comms.run()

        rec_task.run(actor=receiver)
        sen_task.comms = comms
        sen_task.receiver = receiver
        sen_task.run(actor=sender)

        env.run()

        assert env.now == 1.0, "Wrong simulation end time for comms"
        assert receiver.result != "None", "No result for comms"
        assert isinstance(receiver.result, Message), "Wrong result format"
        content = receiver.result.content.data
        assert content["action"] == "move"
        assert content["thought"] == "good"


def test_send_receive_delayed() -> None:
    with EnvironmentContext() as env:
        receiver = ReceiveSend(name="recv")
        sender = ReceiveSend(name="send")

        rec_task = ReceiveTask()
        sen_task = SendTask()

        comms = CommsManager(
            name="Comm",
            send_time=0.25,
            debug_logging=True,
        )
        comms.connect(receiver, "incoming")
        comms.run()

        rec_task.run(actor=receiver)
        sen_task.comms = comms
        sen_task.receiver = receiver
        sen_task.run(actor=sender)

        env.run()

        assert env.now == 1.25, "Wrong simulation end time for comms"
        assert receiver.result != "None", "No result for comms"
        assert isinstance(receiver.result, Message), "Wrong result format"
        content = receiver.result.content.data
        assert content["action"] == "move"
        assert content["thought"] == "good"


def test_degraded() -> None:
    with EnvironmentContext() as env:
        receiver = ReceiveSend(name="recv")
        sender = ReceiveSend(name="send")

        rec_task = ReceiveTask()
        sen_task = SendTask()

        comms = CommsManager(
            name="Comm",
            send_time=0.25,
            debug_logging=True,
        )
        comms.comms_degraded = True
        comms.connect(receiver, "incoming")
        comms.run()

        rec_task.run(actor=receiver)
        sen_task.comms = comms
        sen_task.receiver = receiver
        sen_task.run(actor=sender)

        env.run(until=4)
        comms.comms_degraded = False

        assert receiver.result == "None"


def test_blocked() -> None:
    with EnvironmentContext() as env:
        receiver = ReceiveSend(name="recv")
        sender = ReceiveSend(name="send")

        rec_task = ReceiveTask()
        sen_task = SendTask()

        comms = CommsManager(
            name="Comm",
            send_time=0.25,
            debug_logging=True,
        )
        comms.connect(receiver, "incoming")
        comms.run()

        comms.blocked_links.append((sender, receiver))

        rec_task.run(actor=receiver)
        sen_task.comms = comms
        sen_task.receiver = receiver
        sen_task.run(actor=sender)

        env.run(until=4)
        comms.comms_degraded = False

        assert receiver.result == "None"


def test_comms_wait() -> None:
    with UP.EnvironmentContext() as env:
        store = Store(env=env)
        data_point = []

        def cback(message: MessageContent) -> None:
            data_point.append(message)

        msg = Message(
            sender=UP.Actor(name="me"),
            content=MessageContent(data={"hello": "world"}),
            destination=UP.Actor(name="you"),
        )
        wait_proc = generate_comms_wait(store, cback)
        wait_proc()

        store.put(msg)
        env.run(until=1)
        assert len(data_point) == 1


class Worker(UP.Actor):
    walkie = UP.CommunicationStore(mode="UHF")
    intercom = UP.CommunicationStore(mode="loudspeaker")


def test_worker_talking() -> None:
    with EnvironmentContext() as env:
        w1 = Worker(name="worker1")
        w2 = Worker(name="worker2")

        uhf_comms = CommsManager(name="Walkies", mode="UHF")
        loudspeaker_comms = CommsManager(name="Overhead", mode="loudspeaker")

        uhf_comms.run()
        loudspeaker_comms.run()

        evt1 = uhf_comms.make_put("Hello worker", w1, w2)
        evt2 = loudspeaker_comms.make_put("Hello worker", w2, w1)

        def do() -> SIMPY_GEN:
            yield evt1.as_event()
            yield evt2.as_event()

        env.process(do())

        env.run()
        assert len(w2.walkie.items) == 1
        assert len(w1.intercom.items) == 1



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import pytest

from upstage_des.geography import Spherical


def test_distance(atl: tuple[float, float], nas: tuple[float, float]) -> None:
    dist = Spherical.distance(atl, nas)
    assert dist == pytest.approx(186.94317974521996)


def test_bearing(atl: tuple[float, float], nas: tuple[float, float]) -> None:
    bearing = Spherical.bearing(atl, nas)
    assert bearing == pytest.approx(321.5766379719283)


def test_linspace(atl: tuple[float, float], nas: tuple[float, float]) -> None:
    latlons = Spherical.geo_linspace(atl, nas, 10)
    assert len(latlons) == 11
    lats, lons = zip(*latlons)
    assert pytest.approx(lats[0]) == atl[0]
    assert pytest.approx(lats[-1]) == nas[0]
    assert pytest.approx(lons[0]) == atl[1]
    assert pytest.approx(lons[-1]) == nas[1]


def test_geo_circle(atl: tuple[float, float]) -> None:
    latlons = Spherical.geo_circle(atl, radius=100, num_points=20)
    for lat, lon in latlons:
        d = Spherical.distance(atl, (lat, lon))
        assert d == pytest.approx(100)


def test_point_along(atl: tuple[float, float], nas: tuple[float, float]) -> None:
    pt = Spherical.point_along(atl, nas, 0.0)
    assert pt[0] == atl[0]
    assert pt[1] == atl[1]


def test_bearing_dist_from_point(atl: tuple[float, float], nas: tuple[float, float]) -> None:
    bearing = Spherical.bearing(atl, nas)
    dist = Spherical.distance(atl, nas)
    pt = Spherical.point_from_bearing_dist(atl, bearing, dist, "nmi")
    new_dist = Spherical.distance(pt, nas, "nmi")
    assert new_dist <= 1e-4


def test_cross_track(
    nyc: tuple[float, float], lax: tuple[float, float], atl: tuple[float, float]
) -> None:
    res = Spherical.cross_track_point(nyc, lax, atl)
    dist = Spherical.cross_track_distance(nyc, lax, atl)
    dist2 = Spherical.distance(res, atl)
    assert pytest.approx(dist) == dist2



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""This module contains utility functions.

Note:
    Some of the functions included in this module directly support UPSTAGE's
    other modules, and some are there for the user's convenience.

"""

import inspect
from collections.abc import Sequence
from sys import _getframe as get_frame  # pylint: disable=protected-access
from typing import Any, TypeVar

from .data_types import Location

__all__ = (
    "debug_assert",
    "debug_pause",
    "get_caller_info",
)


SKIP: bool = True


def debug_assert(test: bool, msg: str = "") -> None:
    """Coalesces breakpoints for any failing assert to a single line.

    Coalesces all potential lines where the code may fail into one
    single line that can be marked as a breakpoint.

    Args:
        test (bool): The boolean statement to test, i.e., must evaluate to ``true`` or ``false``.
        msg (str): The message to display if the test is false.

    Note:
        This is necessary because ``pdb`` sometimes does not work well when
        running ``simpy`` due to the way that ``simpy`` handles exceptions.

        This is also helpful when debugging complex behaviors that run the same
        code multiple times.  Instead of manually writing a ``try/except``
        statement, you can use ``debug_assert`` to do that for you, and ignore
        all of them from a single control point.


    Example:
        >>> # 1. Add a ``debug_assert`` statement in your code, e.g.:
        >>> from upstage.utils import debug_assert
        >>> ...
        >>> bar, foo = 0, 1  # <<< change foo to be less than bar to raise
        >>> debug_assert(foo > bar, "foo is not greater than bar")
        >>> # 2. Put a break point on the ``raise error`` line to see why the assert failed.

    """
    if SKIP:
        return

    try:
        assert test, msg
    except AssertionError as error:
        raise error  # <<< ADD BREAKPOINT HERE


def debug_pause(test: bool | None = None) -> None:
    """Call function to pause IDE on debug mode with single breakpoint.

    A helper function to pause the execution of the interpreter when running
    the code from an IDE (e.g., PyCharm).

    Args:
        test (bool, optional): A boolean statement to pause on when it evaluates to ``true``.

    Note:
        This is necessary because ``pdb`` sometimes does not work well when
        running ``simpy`` due to the way that ``simpy`` handles exceptions.

    Note:
        Put a break point on the ``pass`` line to pause the IDE.

    """
    if SKIP:
        return

    if test is not None and test:
        pass  # <<< ADD BREAKPOINT HERE


def get_caller_object(caller_level: int = 2) -> Any:
    """Inspect the stack to see who called you.

    Args:
        caller_level (int, optional): Number of hops up in the stack. Defaults to 2.

    Returns:
        Any: The task object
    """
    try:
        task_frame = inspect.stack()[caller_level]
        task_object = task_frame.frame.f_locals["self"]
        return task_object
    except Exception:
        return None


def get_caller_info(caller_level: int = 1) -> str:
    """Get information from the object that called the function.

    Parameters
    ----------
    caller_level : str, optional
        The number of frames to go back in the call stack.

    """
    try:
        frame = get_frame(caller_level + 1)
        if frame.f_code.co_name == "task":
            try:
                caller = frame.f_code.co_qualname
                if caller:
                    return caller
            except (AttributeError, IndexError):
                ...
        return frame.f_code.co_name
    except ValueError as exc:
        if any("call stack is not deep enough" in arg for arg in exc.args):
            return "Unknown caller"
        raise
    except Exception:
        raise


T = TypeVar("T")


def iterable_convert(item: T | list[T] | tuple[T, ...]) -> list[T]:
    """Convert single objects or tuples into a list.

    Args:
        item (T | list[T] | tuple[T,...]): Object, list, or tuple to convert.

    Returns:
        list[T]: The list version of the input.
    """
    if not isinstance(item, list | tuple):
        return [item]
    return list(item)


def waypoint_time_and_dist(
    start: Location,
    waypoints: Sequence[Location],
    speed: float,
) -> tuple[float, float]:
    """Get the time and distance of a series of locations.

    Args:
        start (Location): Starting point
        waypoints (Sequence[Location]): Waypoints after the start
        speed (float): Travel speed

    Returns:
        tuple[float, float]: Time and Distance over the waypoints.
    """
    dist = 0.0
    current = start
    for wypt in waypoints:
        dist += wypt - current
        current = wypt
    time = dist / speed
    return time, dist



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Geographical math using spherical coordinates for Earth."""

from math import acos, asin, atan, atan2, cos, degrees, radians, sin, sqrt

from upstage_des.math_utils import _vector_dot
from upstage_des.units import unit_convert

from .conversions import SphericalConversions, spherical_radius
from .geo_types import GEO_POINT, LAT_LON, POSITION, POSITIONS, _convert_geo


class Spherical(SphericalConversions):
    """Geographical math using spherical coordinates for Earth."""

    EARTH_RADIUS = spherical_radius  # m

    @staticmethod
    def _bearing(
        lat1: float,
        lon1: float,
        lat2: float,
        lon2: float,
    ) -> float:
        ans = degrees(
            atan2(
                sin(lon2 - lon1) * cos(lat2),
                cos(lat1) * sin(lat2) - sin(lat1) * cos(lat2) * cos(lon2 - lon1),
            )
        )
        ans = ans % 360
        return float(ans)

    @classmethod
    def bearing(
        cls,
        origin: LAT_LON | GEO_POINT,
        destination: LAT_LON | GEO_POINT,
    ) -> float:
        """Calculate the forward bearing from origin to destination.

        Args:
            origin (LAT_LON): Start lat/lon
            destination (LAT_LON): End lat/lon

        Returns:
            float: The forward bearing (0 is north)
        """
        origin = _convert_geo(origin)
        destination = _convert_geo(destination)
        positions = [origin[1], origin[0], destination[1], destination[0]]
        lon1, lat1, lon2, lat2 = map(radians, positions)

        return cls._bearing(lat1, lon1, lat2, lon2)

    @staticmethod
    def _point_along(
        d: float,
        f: float,
        lat1: float,
        lon1: float,
        lat2: float,
        lon2: float,
    ) -> LAT_LON:
        """Get a point a fraction between two points on a great circle.

        Args:
            d (float): Distance between the points (meters)
            f (float): Fraction of the distance for the new point
            lat1 (float): Latitude of the first point (radians)
            lon1 (float): Longitude of the second point (radians)
            lat2 (float): Latitude of the second point (radians)
            lon2 (float): Longitude of the second point (radians)

        Returns:
            LAT_LON: Lat and Lon (degrees)
        """
        a = sin((1 - f) * d) / sin(d)
        b = sin(f * d) / sin(d)
        x = a * cos(lat1) * cos(lon1) + b * cos(lat2) * cos(lon2)
        y = a * cos(lat1) * sin(lon1) + b * cos(lat2) * sin(lon2)
        z = a * sin(lat1) + b * sin(lat2)
        lat = degrees(atan2(z, sqrt(x**2 + y**2)))
        lon = degrees(atan2(y, x))

        return lat, lon

    @classmethod
    def point_along(
        cls,
        origin: LAT_LON | GEO_POINT,
        destination: LAT_LON | GEO_POINT,
        f: float,
    ) -> LAT_LON:
        """Find the location fractionally along a great circle path.

        Args:
            origin (LAT_LON): Lat / Lon (degrees)
            destination (LAT_LON): Lat / Lon (degrees)
            f (float): Fraction along the great circle path to get the point at

        Returns:
            LAT_LON: Point between origin and destination (degrees)
        """
        origin = _convert_geo(origin)
        destination = _convert_geo(destination)
        positions = [origin[1], origin[0], destination[1], destination[0]]
        lon1, lat1, lon2, lat2 = map(radians, positions)
        d = cls.distance(origin, destination, units="m") / spherical_radius
        lat, lon = cls._point_along(d, f, lat1, lon1, lat2, lon2)
        return (lat, lon)

    @classmethod
    def _ll2cart(cls, lat: float, lon: float, radius: float = 1.0) -> POSITION:
        """Latitude and longitude to cartesian on a spherical earth.

        Args:
            lat (float | list[float]): Latitude
            lon (float | list[float]): Longitude
            radius (float, optional): Sphere radius. Defaults to 1.

        Returns:
            POSITION: XYZ coordinates
        """
        theta, lam = radians(lon), radians(lat)
        cl = cos(lam)
        v = (radius * cos(theta) * cl, radius * sin(theta) * cl, radius * sin(lam))
        return v

    @classmethod
    def _cart2lla(cls, cart: POSITION, radius: float = 1.0) -> POSITION:
        """Convert XYZ to lat/lon.

        Args:
            cart (POSITION): Cartesian (XYZ) point.
            radius (float, optional): Radius of the sphere. Defaults to 1.0.

        Returns:
            POSITION: Latitude and longitude (degrees), and altitude (meters)
        """
        a, b, c = cart
        lam = asin(c)
        theta = atan2(b, a)
        alt = sum(x**2 for x in cart) - radius
        return degrees(lam), degrees(theta), alt

    @classmethod
    def geo_linspace(
        cls,
        start: LAT_LON | GEO_POINT,
        end: LAT_LON | GEO_POINT,
        num_segments: int = 10,
    ) -> list[LAT_LON]:
        """Make a discrete great circle path between two points.

        Altitudes are not considered since they do not affect angular coordinates.

        The number of points is the number

        Args:
            start (LAT_LON): Start point
            end (LAT_LON): End point
            num_segments (int, optional): Number of segments on the path. Defaults to 10.

        Returns:
           list[LAT_LON]: Lattiude and Longitude (degrees)
        """
        start = _convert_geo(start)
        end = _convert_geo(end)
        lats, lons, _ = cls.geo_linspace_with_ecef(start, end, num_segments)
        latlon: list[LAT_LON] = [(la, lo) for la, lo in zip(lats, lons)]
        return latlon

    @classmethod
    def geo_linspace_with_ecef(
        cls,
        start: LAT_LON | GEO_POINT,
        end: LAT_LON | GEO_POINT,
        num_segments: int = 10,
    ) -> tuple[list[float], list[float], POSITIONS]:
        """Make a discrete great circle path between two points.

        Altitudes are not considered since they do not affect angular coordinates.

        Latitude and Longitude are all in degrees

        Args:
            start (LAT_LON): Starting lat/lon (degrees)
            end (LAT_LON): Ending lat/lon (degrees)
            num_segments (int, optional): Number of segments to make. Defaults to 10.

        Returns:
            tuple[list[float], list[float], POSITIONS]: Latitude, Longitude, and ECEF points.
        """
        start = _convert_geo(start)
        end = _convert_geo(end)
        if num_segments < 2:
            raise ValueError("Not enough segments for interpolation")
        # We keep the earth radius to be 1 since we're interpolating along radians only
        v1 = cls._ll2cart(start[0], start[1])
        v2 = cls._ll2cart(end[0], end[1])
        # solve for the parameter value that equals the end point
        d = _vector_dot(v1, v2)
        common = sqrt(-1 / ((d - 1) * (d + 1)))
        alpha, beta = -d * common, common
        a, _, _ = v1
        d, _, _ = v2
        lead = a * alpha + beta * d
        sqrt_inner = a**2 * alpha**2 + a**2 + 2 * a * alpha * beta * d + beta**2 * d**2 - d**2
        if sqrt_inner < 0:
            raise ValueError("geo_linspace fails due to negative sqrt")
        t = 2 * atan((lead + sqrt(sqrt_inner)) / (a + d))
        W = tuple([alpha * x + beta * y for x, y in zip(v1, v2)])
        assert len(W) == 3
        W = (W[0], W[1], W[2])
        # test that the t is right
        ct, st = cos(t), sin(t)
        end_pt = tuple([ct * x + st * w for x, w in zip(v1, W)])
        assert len(end_pt) == 3
        end_pt = (end_pt[0], end_pt[1], end_pt[2])
        is_close = all(abs(a - b) <= 1e-8 for a, b in zip(end_pt, v2))
        if not is_close:
            t = 2 * atan((lead - sqrt(sqrt_inner)) / (a + d))

        ts = [(t / num_segments) * i for i in range(0, num_segments + 1)]
        v_cos = [tuple(cos(t) * v for v in v1) for t in ts]
        w_sin = [tuple(sin(t) * w for w in W) for t in ts]
        end_points: list[POSITION] = []
        for v, w in zip(v_cos, w_sin):
            pt = (v[0] + w[0], v[1] + w[1], v[2] + w[2])
            end_points.append(pt)

        lla = cls.ecef2lla(end_points)

        return [lat for lat, *_ in lla], [lon for _, lon, _ in lla], end_points

    @classmethod
    def geo_circle(
        cls,
        center: LAT_LON | GEO_POINT,
        radius: float,
        radius_units: str = "nmi",
        num_points: int = 10,
    ) -> list[LAT_LON]:
        """Make an approximate circle by sweeping bearing from a central location.

        Altitudes are not considered since they do not affect angular coordinates.

        Args:
            center (LAT_LON): Center point (degrees)
            radius (float): Radius of the circle
            radius_units (str, optional): Units of the radius. Defaults to 'nmi'.
            num_points (int, optional): Number of points on the circle. Defaults to 10.

        Returns:
            list[LAT_LON]: Latitude and Longitude (degrees) of the circle
        """
        center = _convert_geo(center)
        lats: list[float] = []
        lons: list[float] = []
        for fraction in range(0, num_points + 1):
            f = fraction / num_points
            bearing = f * 360
            lat, lon = cls.point_from_bearing_dist(
                center,
                bearing,
                radius,
                distance_units=radius_units,
            )
            lats.append(lat)
            lons.append(lon)

        latlon: list[LAT_LON] = [(la, lo) for la, lo in zip(lats, lons)]
        return latlon

    @classmethod
    def distance(
        cls,
        loc1: LAT_LON | GEO_POINT,
        loc2: LAT_LON | GEO_POINT,
        units: str = "nmi",
    ) -> float:
        """Find distance between two points.

        Return the great circle distance from the loc1 to another loc2.

        Args:
            loc1 (LAT_LON): Starting point (degrees)
            loc2 (LAT_LON): Ending point (degrees)
            units (str, optional): Units requested for distance. Defaults to 'nmi'.

        Returns:
            float: Distances in the defined units
        """
        loc1 = _convert_geo(loc1)
        loc2 = _convert_geo(loc2)
        positions = [loc1[1], loc1[0], loc2[1], loc2[0]]
        lon1, lat1, lon2, lat2 = map(radians, positions)
        a = sin(0.5 * (lat2 - lat1)) ** 2 + cos(lat1) * cos(lat2) * sin(0.5 * (lon2 - lon1)) ** 2

        dist_m = spherical_radius * 2 * atan2(sqrt(a), sqrt(1.0 - a))
        return unit_convert(dist_m, "m", units)

    @classmethod
    def distance_and_bearing(
        cls,
        loc1: LAT_LON | GEO_POINT,
        loc2: LAT_LON | GEO_POINT,
        units: str = "nmi",
    ) -> tuple[float, float]:
        """Find great circle distance and forward bearing between two points.

        Args:
            loc1 (LAT_LON): Starting point
            loc2 (LAT_LON): Ending point
            units (str, optional): Distance units requested. Defaults to 'nmi'.

        Returns:
            tuple[float, float]: Distance and Bearing
        """
        dist = cls.distance(loc1, loc2, units=units)
        bearing = cls.bearing(loc1, loc2)
        return dist, bearing

    @classmethod
    def point_from_bearing_dist(
        cls,
        point: LAT_LON | GEO_POINT,
        bearing: float,
        distance: float,
        distance_units: str = "nmi",
    ) -> LAT_LON:
        """Get a Location from a starting point, a bearing, and a distance.

        Args:
            point (LAT_LON): Starting point (degrees)
            bearing (float): Bearing to travel along (degrees)
            distance (float): Distance to travel
            distance_units (str, optional): Units of the distance. Defaults to 'nmi'.

        Returns:
            LAT_LON: The point
        """
        point = _convert_geo(point)
        bearing = radians(bearing)
        dist = unit_convert(distance, distance_units, "m")

        lat, lon = point
        lat1 = radians(lat)
        lon1 = radians(lon)

        lat2 = asin(
            sin(lat1) * cos(dist / spherical_radius)
            + cos(lat1) * sin(dist / spherical_radius) * cos(bearing)
        )

        lon2 = lon1 + atan2(
            sin(bearing) * sin(dist / spherical_radius) * cos(lat1),
            cos(dist / spherical_radius) - sin(lat1) * sin(lat2),
        )

        return degrees(lat2), degrees(lon2)

    @classmethod
    def cross_track_distance(
        cls,
        origin: LAT_LON | GEO_POINT,
        destination: LAT_LON | GEO_POINT,
        point: LAT_LON | GEO_POINT,
        units: str = "nmi",
    ) -> float:
        """Find the minimum distance from a point to a great circle path.

        Args:
            origin (LAT_LON): Great circle start
            destination (LAT_LON): Great circle end
            point (LAT_LON): Point to start distance measurement
            units (str, optional): Units for the distance output. Defaults to 'nmi'.

        Returns:
            float: Distance from point to origin->destination great circle
        """
        origin = _convert_geo(origin)
        destination = _convert_geo(destination)
        point = _convert_geo(point)
        delta_1_3 = cls.distance(origin, point, units="m") / spherical_radius
        theta_1_3 = radians(cls.bearing(origin, point))
        theta_1_2 = radians(cls.bearing(origin, destination))

        dist = asin(sin(delta_1_3) * sin(theta_1_3 - theta_1_2)) * spherical_radius
        return unit_convert(abs(dist), "m", units)

    @classmethod
    def cross_track_point(
        cls,
        origin: LAT_LON | GEO_POINT,
        destination: LAT_LON | GEO_POINT,
        point: LAT_LON | GEO_POINT,
    ) -> LAT_LON:
        """Find the point on a great circle that minimizes the distance to another point.

        Args:
            origin (LAT_LON): Great circle start
            destination (LAT_LON): Great circle end
            point (LAT_LON): Point to start distance measurement

        Returns:
            LAT_LON: Location on great circle closest to a point
        """
        origin = _convert_geo(origin)
        destination = _convert_geo(destination)
        point = _convert_geo(point)
        delta_1_3 = cls.distance(origin, point, units="m") / spherical_radius
        theta_1_3 = radians(cls.bearing(origin, point))
        theta_1_2 = radians(cls.bearing(origin, destination))
        d = asin(sin(delta_1_3) * sin(theta_1_3 - theta_1_2))

        dist_from_origin = acos(cos(delta_1_3) / cos(d)) * spherical_radius
        cross_track_point = cls.point_from_bearing_dist(
            origin,
            degrees(theta_1_2),
            dist_from_origin,
            "m",
        )
        return cross_track_point

    @classmethod
    def ecef_linspace(
        cls,
        start: LAT_LON | GEO_POINT,
        finish: LAT_LON | GEO_POINT,
        start_alt: float,
        finish_alt: float,
        segments: int,
    ) -> POSITIONS:
        """Get an array of spaced points along a greate circle path in ECEF.

        Args:
            start (LAT_LON): Start point (lat/lon degrees)
            finish (LAT_LON): End point (lat/lon degrees)
            start_alt (float): Starting altitude (meters)
            finish_alt (float): Ending altitude (meters)
            segments (int): Number of segments to create

        Returns:
            POSITIONS: ECEF coordinates
        """
        start = _convert_geo(start)
        finish = _convert_geo(finish)
        lats, lons = cls.geo_linspace(start, finish, num_segments=segments)
        delta_alt = finish_alt - start_alt
        alts = [start_alt + delta_alt * i / segments for i in range(segments + 1)]
        lla = [(lat, lon, alt) for lat, lon, alt in zip(lats, lons, alts)]
        return cls.lla2ecef(lla)

    @classmethod
    def ecef_and_geo_linspace(
        cls,
        start: LAT_LON | GEO_POINT,
        finish: LAT_LON | GEO_POINT,
        start_alt: float,
        finish_alt: float,
        segments: int,
    ) -> tuple[POSITIONS, POSITIONS]:
        """Get an array of spaced points along a greate circle path in ECEF and LLA.

        Args:
            start (LAT_LON): Start point (lat/lon degrees)
            finish (LAT_LON): End point (lat/lon degrees)
            start_alt (float): Starting altitude (meters)
            finish_alt (float): Ending altitude (meters)
            segments (int): Number of segments to create

        Returns:
            POSITIONS: ECEF coordinates
            POSITIONS: Lat/Lon/Alt (degrees)
        """
        start = _convert_geo(start)
        finish = _convert_geo(finish)
        lats, lons, _ = cls.geo_linspace_with_ecef(start, finish, num_segments=segments)
        delta_alt = finish_alt - start_alt
        alts = [start_alt + delta_alt * i / segments for i in range(segments + 1)]
        lla = [(lat, lon, alt) for lat, lon, alt in zip(lats, lons, alts)]
        return cls.lla2ecef(lla), lla



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

import random

import pytest


@pytest.fixture()
def atl() -> tuple[float, float]:
    return (33.7490, -84.3880)


@pytest.fixture()
def nas() -> tuple[float, float]:
    return (36.1627, -86.7816)


@pytest.fixture()
def nyc() -> tuple[float, float]:
    return (40.7128, -74.0060)


@pytest.fixture()
def lax() -> tuple[float, float]:
    return (34.0522, -118.2437)


@pytest.fixture()
def tall() -> tuple[float, float]:
    return (30.4383, -84.2807)


def randvals(rows: int, cols: int) -> list[list[float]]:
    ans: list[list[float]] = []
    for _ in range(rows):
        v = [random.random() for _ in range(cols)]
        ans.append(v)
    return ans


@pytest.fixture()
def random_lla() -> list[tuple[float, float, float]]:
    lla = randvals(10, 3)
    lat, lon, alt = zip(*lla)
    lat = [-90 + 180 * la for la in lat]
    lon = [-180 + 360 * lo for lo in lon]
    alt = [a * 10_000 for a in alt]
    return [(a, b, c) for a, b, c in zip(lat, lon, alt)]


POS = tuple[float, float, float]


@pytest.fixture(
    params=[
        (0, 0, 0, 100, "nmi", ["ENTER", "EXIT"]),
        (5000, 5000, 0, 100, "nmi", ["ENTER", "EXIT"]),
        (0, 0, 0, 190, "nmi", ["START_INSIDE", "EXIT"]),
        (5000, 5000, 0, 190, "nmi", ["START_INSIDE", "EXIT"]),
        (0, 0, 0, 220, "nmi", ["START_INSIDE", "END_INSIDE"]),
        (5000, 5000, 0, 220, "nmi", ["START_INSIDE", "END_INSIDE"]),
    ],
)
def intersect_positions(
    nas: tuple[float, float],
    tall: tuple[float, float],
    atl: tuple[float, float],
    request: pytest.FixtureRequest,
) -> tuple[tuple[POS, POS, POS], float, str, list[str]]:
    # degrees and meters
    parm: tuple[float, float, float, float, str, list[str]] = request.param
    start_alt, finish_alt, sensor_alt, sensor_range, range_units, answer = parm

    start_lla = (nas[0], nas[1], start_alt)
    finish_lla = (tall[0], tall[1], finish_alt)
    sensor_lla = (atl[0], atl[1], sensor_alt)
    return (start_lla, finish_lla, sensor_lla), sensor_range, range_units, answer


@pytest.fixture
def short_intersections() -> tuple[tuple[POS, POS, POS], float, str]:
    start_lla = (61.10051577739581, -154.64858364056806, 100.0)
    finish_lla = (61.15466234550906, -154.77763243723408, 100.0)
    sensor_lla = (58.67779481, -154.11651336, 0.0)
    sensor_radius = 277799.8988808368
    return (start_lla, finish_lla, sensor_lla), sensor_radius, "m"



# Copyright (C) 2025 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Tests for a bug where bad queue order keeps Monitoring*Stores from working."""

from simpy import Container, Environment, FilterStore, Store

from upstage_des.base import EnvironmentContext
from upstage_des.events import Get, Put
from upstage_des.resources.monitoring import (
    SelfMonitoringContainer,
    SelfMonitoringFilterStore,
    SelfMonitoringSortedFilterStore,
    SelfMonitoringStore,
)
from upstage_des.type_help import SIMPY_GEN


def _container_check(container: Container, env: Environment) -> dict[str, tuple[float, float]]:
    """Run a procedure against a container."""
    data: dict[str, tuple[float, float]] = {}

    def _get_proc() -> SIMPY_GEN:
        yield env.timeout(1.5)
        yield Get(container, 1.0).as_event()
        data["one"] = (env.now, 1.0)

    def _put_one() -> SIMPY_GEN:
        yield env.timeout(1)
        yield Put(container, 1.0).as_event()
        data["put one"] = (env.now, 1.0)

    def _put_two() -> SIMPY_GEN:
        yield env.timeout(1)
        yield Put(container, 0.8).as_event()
        data["put two"] = (env.now, 1.0)

    env.process(_get_proc())
    env.process(_put_one())
    env.process(_put_two())

    env.run()

    return data


def test_monitoring_container_get() -> None:
    # The container outputs "True" if a put is successful (to the trigger),
    # so our original failure to output True should cause a successful
    # follow-on put to not happen.

    with EnvironmentContext() as env:
        smc = Container(env, init=1.1, capacity=2)
        data = _container_check(smc, env)

        assert data.get("one", 0.0) == (1.5, 1.0)
        assert data.get("put one", 0.0) == (1.5, 1.0)
        assert data.get("put two", 0.0) == (1.5, 1.0)

    with EnvironmentContext() as env:
        smc = SelfMonitoringContainer(env, init=1.1, capacity=2)
        data2 = _container_check(smc, env)

        assert data2 == data


def _store_process(
    store: Store, env: Environment, filter: bool = True
) -> dict[str, tuple[float, int] | list[int]]:
    """Run a filtering store process."""
    store.items.append(2)

    data: dict[str, tuple[float, int] | list[int]] = {}

    def _proc_one() -> SIMPY_GEN:
        if filter:
            res = yield Get(store, filter=lambda x: x == 3).as_event()
        else:
            res = yield Get(store).as_event()
        data["one"] = (env.now, res)

    def _proc_two() -> SIMPY_GEN:
        if filter:
            res = yield Get(store, filter=lambda x: x == 4).as_event()
        else:
            res = yield Get(store).as_event()
        data["two"] = (env.now, res)

    # The bug is that if you stack the requests in an order where the
    # first one fails, it's not going to succeed later
    def _proc_three() -> SIMPY_GEN:
        yield env.timeout(2.0)
        yield Put(store, 4).as_event()
        yield env.timeout(0.5)
        yield Put(store, 3).as_event()

    env.process(_proc_one())
    env.process(_proc_two())
    env.process(_proc_three())

    env.run()

    data["final"] = list(store.items)
    return data


def test_monitoring_store_get() -> None:
    # This should not be an issue because simpy Store _do_get|put always
    # returns None
    with EnvironmentContext() as env:
        smst = SelfMonitoringStore(env)
        data = _store_process(smst, env, filter=False)
        assert data.get("one", 0.0) == (0.0, 2)
        assert data.get("two", 0.0) == (2.0, 4)
        assert smst.items == [3]


def test_monitoring_filter_store_get() -> None:
    with EnvironmentContext() as env:
        smfst = FilterStore(env)
        data = _store_process(smfst, env)
        assert data.get("one", 0.0) == (2.5, 3)
        assert data.get("two", 0.0) == (2.0, 4)
        assert data.get("final", [1]) == [2]

    with EnvironmentContext() as env:
        smfst = SelfMonitoringFilterStore(env)
        data2 = _store_process(smfst, env)
        assert data2 == data


def test_monitoring_sorted_filter_store_get() -> None:
    with EnvironmentContext() as env:
        smsfst = SelfMonitoringSortedFilterStore(env)

        data = _store_process(smsfst, env)
        assert data.get("one", 0.0) == (2.5, 3)
        assert data.get("two", 0.0) == (2.0, 4)
        assert data.get("final", [1]) == [2]


if __name__ == "__main__":
    test_monitoring_container_get()



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""This file contains a Container that allows reservations."""

from collections.abc import Generator
from typing import Any

from simpy import Environment, Event

from ..task import process

__all__ = ("ReserveContainer",)


class ReserveContainer:
    """A store that allows requests to be scheduled in advance.

    This is not a true container (you can't yield on a reserve slot!).
    """

    def __init__(
        self,
        env: Environment,
        init: float = 0.0,
        capacity: float = float("inf"),
    ) -> None:
        """Create a container-like object that allows reservations.

        Note that this container doesn't actually yield to SimPy when requesting.

        Use it to determine if anything is avaiable for reservation, but there is no
        queue for getting a reservation.

        Args:
            env (Environment): The SimPy Environment
            init (float, optional): Initial amount available. Defaults to 0.0.
            capacity (float, optional): Total capacity. Defaults to float("inf").
        """
        self.capacity = capacity
        self._env = env
        self._level = init
        self._real_level = init
        self._queued: dict[Any, tuple[float, float]] = {}

    @property
    def remaining(self) -> float:
        """Return the amount remaining in the container.

        Returns:
            float: Amount remaining
        """
        return self._level

    @property
    def available(self) -> float:
        """Return the amount remaining in the container.

        Returns:
            float: Amount remaining.
        """
        return self.remaining

    @property
    def queued(self) -> list[Any]:
        """Get the queued requesters.

        Returns:
            list[Any]: List of requesters.
        """
        return list(self._queued.keys())

    @process
    def _expire_request(self, requester: Any, time: float) -> Generator[Event, None, None]:
        """Expire the request after an expiration period or at a specific time.

        :param request: the Request namedtuple object
        :param expiration: the expiration Event object

        :type request:
        :type expiration:

        """
        yield self._env.timeout(time)
        self.cancel_request(requester)

    def reserve(self, requester: Any, quantity: float, expiration: float | None = None) -> bool:
        """Reserve a quantity of storage."""
        if self.available < quantity:
            return False
        elif requester not in self._queued:
            self._level -= quantity
            self._queued[requester] = (quantity, self._env.now)
            if expiration is not None:
                self._expire_request(requester, expiration)
            return True
        else:
            return False

    def cancel_request(self, requester: Any) -> bool:
        """Have a request cancelled."""
        if requester not in self._queued:
            return False
        else:
            request = [x for x in self._queued if x is requester]
            if not request:
                raise ValueError("Requester not available to cancel")
            self._level += self._queued[requester][0]
            self._queued.pop(request[0])
            return True

    def take(self, requester: Any) -> float:
        """If in queue, allow requester take the actual quantity."""
        if requester not in self._queued:
            raise ValueError("Requester is not in queue, cannot take.")
        else:
            amt, _ = self._queued.pop(requester)
            self._real_level -= amt
            return amt

    def put(self, amount: float, capacity_increase: bool = False) -> None:
        """Put some quantity back in."""
        new = self._level + amount
        if new > self.capacity and not capacity_increase:
            raise ValueError("Adding too much.")
        else:
            self._level = new
            self._real_level += amount



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from typing import Protocol

import pytest

import upstage_des.api as UP
from upstage_des.motion import SteppedMotionManager
from upstage_des.type_help import TASK_GEN
from upstage_des.utils import waypoint_time_and_dist


class LocatedActor(Protocol):
    location: UP.CartesianLocation


class StaticSensor(UP.Actor):
    radius = UP.State[float](valid_types=float, default=2.0)
    history = UP.State[list](default_factory=list)
    location = UP.State[UP.CartesianLocation](valid_types=(UP.CartesianLocation,))

    def entity_entered_range(self, detected: LocatedActor) -> None:
        dist = self.location - detected.location
        self.history.append((self.env.now, "saw", detected.location, dist))

    def entity_exited_range(self, detected: LocatedActor) -> None:
        dist = self.location - detected.location
        self.history.append((self.env.now, "lost", detected.location, dist))


class Mover(UP.Actor):
    location = UP.CartesianLocationChangingState()
    history = UP.State[list](default_factory=list, recording=False)
    radius = UP.State[float](default=2.0)
    visible = UP.DetectabilityState(default=True)
    speed = UP.State[float](default=1.0)

    def entity_entered_range(self, detected: LocatedActor) -> None:
        dist = self.location - detected.location
        self.history.append((self.env.now, "saw", detected, dist))

    def entity_exited_range(self, detected: LocatedActor) -> None:
        dist = self.location - detected.location
        self.history.append((self.env.now, "lost", detected, dist))


class DoMotion(UP.Task):
    waypoints: list[UP.CartesianLocation]

    # Hacking together, assuming start location is (2, 1, 0)
    def task(self, *, actor: Mover) -> TASK_GEN:
        waypoints = [x.copy() for x in self.waypoints]
        time, dist = waypoint_time_and_dist(actor.location, waypoints, actor.speed)
        actor.activate_state(
            state="location",
            task=self,
            speed=actor.speed,
            waypoints=waypoints,
        )
        yield UP.Wait(time)
        actor.deactivate_all_states(task=self)


class EndDetectable(UP.Task):
    time: float

    def task(self, *, actor: Mover) -> TASK_GEN:
        yield UP.Wait(self.time)
        actor.visible = False


class StartDetectable(UP.Task):
    time: float

    def task(self, *, actor: Mover) -> TASK_GEN:
        yield UP.Wait(self.time)
        actor.visible = True


def test_basic_functions() -> None:
    with UP.EnvironmentContext() as env:
        motion = SteppedMotionManager(0.01)
        UP.add_stage_variable("motion_manager", motion)

        sense = StaticSensor(
            name="Static",
            radius=1.7,
            location=UP.CartesianLocation(0, 0, 0),
        )
        move = Mover(
            name="A_Mover",
            location=UP.CartesianLocation(2, 1, 0),
        )

        motion.add_sensor(sense, "radius", "location")
        # This line is handled automatically now
        # motion.add_detectable(move, "location")

        waypoints = [
            UP.CartesianLocation(1, 1, 1),
            UP.CartesianLocation(0, 0, 2),
        ]

        move_task = DoMotion()
        move_task.waypoints = waypoints
        move_task.run(actor=move)
        motion.run()
        env.run()
        assert len(sense.history[0]) == 4, "Wrong intersection history"


def test_dual_sense() -> None:
    with UP.EnvironmentContext() as env:
        motion = SteppedMotionManager(0.01)
        UP.add_stage_variable("motion_manager", motion)

        move1 = Mover(
            name="A_Mover1",
            location=UP.CartesianLocation(-3, 0, 1),
            radius=2,
        )
        move2 = Mover(
            name="A_Mover2",
            location=UP.CartesianLocation(3, 0, 0.5),
            radius=1,
        )

        for obj in [move1, move2]:
            motion.add_sensor(obj, "radius", "location")
            motion.add_detectable(obj, "location")

        waypoints1 = [
            UP.CartesianLocation(3, 0, 0.5),
        ]
        waypoints2 = [
            UP.CartesianLocation(-3, 0, 1),
        ]

        move_task1 = DoMotion()
        move_task1.waypoints = waypoints1
        move_task1.run(actor=move1)

        move_task2 = DoMotion()
        move_task2.waypoints = waypoints2
        move_task2.run(actor=move2)

        motion.run()
        env.run()
        assert len(move1.history[0]) == 4, "Wrong intersection history"
        assert len(move2.history[0]) == 4, "Wrong intersection history"
        assert move1.history[0][2] is move2
        assert move1.history[1][2] is move2
        assert move2.history[0][2] is move1
        assert move2.history[1][2] is move1
        # different distances
        assert move1.history[0][3] != move2.history[0][3]
        # but correct distances - within a loose tolerance
        assert pytest.approx(move1.history[0][3], abs=0.1) == 2
        assert pytest.approx(move2.history[0][3], abs=0.1) == 1


def test_detectability_change() -> None:
    with UP.EnvironmentContext() as env:
        motion = SteppedMotionManager(0.01)
        UP.add_stage_variable("motion_manager", motion)

        move1 = Mover(
            name="A_Mover1",
            location=UP.CartesianLocation(-3, 0, 1),
            radius=2,
        )
        move2 = Mover(
            name="A_Mover2",
            location=UP.CartesianLocation(3, 0, 0.5),
            radius=1,
        )

        for obj in [move1, move2]:
            motion.add_sensor(obj, "radius", "location")
            motion.add_detectable(obj, "location")

        waypoints1 = [
            UP.CartesianLocation(3, 0, 0.5),
        ]
        waypoints2 = [
            UP.CartesianLocation(-3, 0, 1),
        ]

        move_task1 = DoMotion()
        move_task1.waypoints = waypoints1
        move_task1.run(actor=move1)

        move_task2 = DoMotion()
        move_task2.waypoints = waypoints2
        move_task2.run(actor=move2)

        task_end = EndDetectable()
        task_end.time = 2.9
        task_end.run(actor=move1)

        task_start = StartDetectable()
        task_start.time = 3.1
        task_start.run(actor=move1)

        motion.run()
        env.run()
        assert len(move1.history[0]) == 4, "Wrong intersection history"
        assert len(move2.history[0]) == 4, "Wrong intersection history"
        assert move1.history[0][2] is move2
        assert move1.history[1][2] is move2
        assert move2.history[0][2] is move1
        assert move2.history[1][2] is move1
        assert move2.history[1][0] == 2.9
        assert move2.history[2][0] == 3.1
        assert pytest.approx(move2.history[2][0], abs=0.01) == 3.1


if __name__ == "__main__":
    test_basic_functions()



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""The elements in the UPSTAGE Application Programmable Interface."""

# Core
# Director, stage, and Exceptions
# Actor
from upstage_des.actor import Actor
from upstage_des.base import (
    EnvironmentContext,
    MotionAndDetectionError,
    NamedUpstageEntity,
    RulesError,
    SimulationError,
    UpstageBase,
    UpstageError,
    add_stage_variable,
    get_stage,
    get_stage_variable,
)

# Comms
from upstage_des.communications.comms import CommsManager, Message, MessageContent

# Constants
from upstage_des.constants import PLANNING_FACTOR_OBJECT

# Data types
from upstage_des.data_types import (
    CartesianLocation,
    CartesianLocationData,
    GeodeticLocation,
    GeodeticLocationData,
    Location,
)

# Events
from upstage_des.events import All, Any, Event, FilterGet, Get, Put, ResourceHold, Wait

# Motion
from upstage_des.motion import SensorMotionManager, SteppedMotionManager

# Task network nucleus
from upstage_des.nucleus import NucleusInterrupt, TaskNetworkNucleus

# Resources
from upstage_des.resources.container import (
    ContainerEmptyError,
    ContainerError,
    ContainerFullError,
    ContinuousContainer,
)
from upstage_des.resources.monitoring import (
    SelfMonitoringContainer,
    SelfMonitoringContinuousContainer,
    SelfMonitoringFilterStore,
    SelfMonitoringReserveContainer,
    SelfMonitoringSortedFilterStore,
    SelfMonitoringStore,
)
from upstage_des.resources.reserve import ReserveContainer
from upstage_des.resources.sorted import SortedFilterGet, SortedFilterStore

# Nucleus-friendly states
from upstage_des.state_sharing import SharedLinearChangingState

# States
from upstage_des.states import (
    CartesianLocationChangingState,
    CommunicationStore,
    DetectabilityState,
    GeodeticLocationChangingState,
    LinearChangingState,
    ResourceState,
    State,
)

# Task
from upstage_des.task import DecisionTask, InterruptStates, Task, TerminalTask, process

# Task Networks
from upstage_des.task_network import TaskLinks, TaskNetwork, TaskNetworkFactory

# Conversion
from upstage_des.units import unit_convert

__all__ = [
    "UpstageError",
    "SimulationError",
    "MotionAndDetectionError",
    "RulesError",
    "Actor",
    "PLANNING_FACTOR_OBJECT",
    "UpstageBase",
    "NamedUpstageEntity",
    "EnvironmentContext",
    "add_stage_variable",
    "get_stage_variable",
    "get_stage",
    "All",
    "Any",
    "Event",
    "Get",
    "FilterGet",
    "SortedFilterGet",
    "Put",
    "ResourceHold",
    "Wait",
    "ContainerEmptyError",
    "ContainerError",
    "ContainerFullError",
    "ContinuousContainer",
    "SelfMonitoringContainer",
    "SelfMonitoringContinuousContainer",
    "SelfMonitoringFilterStore",
    "SelfMonitoringSortedFilterStore",
    "SelfMonitoringReserveContainer",
    "SelfMonitoringStore",
    "ReserveContainer",
    "SortedFilterStore",
    "CartesianLocation",
    "GeodeticLocation",
    "Location",
    "CartesianLocationData",
    "GeodeticLocationData",
    "LinearChangingState",
    "CartesianLocationChangingState",
    "State",
    "GeodeticLocationChangingState",
    "DetectabilityState",
    "ResourceState",
    "CommunicationStore",
    "DecisionTask",
    "Task",
    "process",
    "InterruptStates",
    "TerminalTask",
    "TaskNetwork",
    "TaskNetworkFactory",
    "TaskLinks",
    "TaskNetworkNucleus",
    "NucleusInterrupt",
    "SharedLinearChangingState",
    "CommsManager",
    "Message",
    "MessageContent",
    "SensorMotionManager",
    "SteppedMotionManager",
    "unit_convert",
]



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""This sub-module contains advanced stores and containers for UPSTAGE."""

from .container import (
    ContainerEmptyError,
    ContainerError,
    ContainerFullError,
    ContinuousContainer,
)
from .monitoring import (
    SelfMonitoringContainer,
    SelfMonitoringContinuousContainer,
    SelfMonitoringFilterStore,
    SelfMonitoringReserveContainer,
    SelfMonitoringSortedFilterStore,
    SelfMonitoringStore,
)
from .reserve import ReserveContainer
from .sorted import SortedFilterGet, SortedFilterStore

__all__ = [
    "ContinuousContainer",
    "ContainerEmptyError",
    "ContainerError",
    "ContainerFullError",
    "SelfMonitoringStore",
    "SelfMonitoringFilterStore",
    "SelfMonitoringContainer",
    "SelfMonitoringContinuousContainer",
    "SelfMonitoringSortedFilterStore",
    "SelfMonitoringReserveContainer",
    "ReserveContainer",
    "SortedFilterStore",
    "SortedFilterGet",
]



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)
# Licensed under the 3-Clause BSD License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
import multiprocessing as mp

import pytest
import simpy as SIM

from upstage_des.base import (
    STAGE_CONTEXT_VAR,
    EnvironmentContext,
    NamedUpstageEntity,
    UpstageBase,
    UpstageError,
    add_stage_variable,
)


def test_context() -> None:
    with EnvironmentContext() as env:
        assert isinstance(env, SIM.Environment)
        env.run(until=3)
        assert env.now == 3

    with EnvironmentContext(initial_time=3.1) as env:
        assert isinstance(env, SIM.Environment)
        assert env.now == 3.1


def test_entity_naming() -> None:
    class TestItem(NamedUpstageEntity): ...

    class Example(TestItem, entity_groups=["This"]): ...

    class Ignorable(NamedUpstageEntity, add_to_entity_groups=False): ...

    # This shows how the entity group ignorance doesn't cascade down.
    class Ignorable2(Ignorable): ...

    with pytest.warns(UserWarning, match="Environment not created*"):
        other = Example()

    with pytest.raises(UpstageError):
        other.env

    with EnvironmentContext() as env:
        # Note that due to the context, we can access the environment
        # even if we couldn't before.
        assert env is other.env

        t = TestItem()
        t2 = TestItem()
        e = Example()
        _ = Ignorable()
        _ = Ignorable2()

        items = t.get_entity_group("TestItem")
        assert len(items) == 3
        assert t in items
        assert t2 in items
        assert e in items

        items = t.get_entity_group("This")
        assert len(items) == 1
        assert e in items

        items = t.get_entity_group("Example")
        assert len(items) == 1
        assert e in items

        items = t.get_entity_group("Ignorable")
        assert len(items) == 0

        items = t.get_entity_group("Ignorable2")
        assert len(items) == 1


def test_stage() -> None:
    with EnvironmentContext():
        add_stage_variable("A variable", 3.14)
        ans = STAGE_CONTEXT_VAR.get()
        assert ans.get("A variable", 0.1) == 3.14
        assert ans.get("random") is not None
        assert len(ans) == 2
        with pytest.raises(UpstageError):
            add_stage_variable("A variable", 2)


def test_random() -> None:
    with EnvironmentContext(random_seed=1234986):
        cl = UpstageBase()
        num = cl.stage.random.uniform(1, 3)
        assert pytest.approx(num) == 2.348057489610457

    from random import Random

    rng = Random(1234986)
    with EnvironmentContext(random_gen=rng):
        cl = UpstageBase()
        num = cl.stage.random.uniform(1, 3)
        assert pytest.approx(num) == 2.348057489610457

    with EnvironmentContext():
        num = cl.stage.random.uniform(1, 3)


def a_simulation(t: float) -> float:
    with EnvironmentContext() as env:
        env.run(until=env.now + t)
        return env.now


def test_multiproc_stability() -> None:
    inputs = [1.2, 3.4, 5.6, 10.11, 74.31]

    with mp.Pool(3) as pool:
        res = pool.map(a_simulation, inputs)

    assert res == inputs


@pytest.mark.parametrize(
    ["unit", "mult"],
    [
        ("min", 60),
        ("Minutes", 60),
        ("s", 3600),
        ("second", 3600),
        ("hours", 1),
        ("hr", 1),
        (None, 1),
    ],
)
def test_pretty_times(unit: str, mult: int) -> None:
    """Test that if we do time in regular ways, we get standard logging."""
    times_in_hours = [3, 28, 24 * 15 + 6.5]
    times_in_hours = [x * mult for x in times_in_hours]
    with EnvironmentContext(initial_time=times_in_hours[0]) as env:
        add_stage_variable("time_unit", unit)
        base = UpstageBase()
        assert base.pretty_now == "[Day    0 - 03:00:00]"
        env.run(until=times_in_hours[1])
        assert base.pretty_now == "[Day    1 - 04:00:00]"
        env.run(until=times_in_hours[2])
        assert base.pretty_now == "[Day   15 - 06:30:00]"


@pytest.mark.parametrize("unit", ["ticks", "week", "day", "microseconds"])
def test_pretty_time_nonstandard(unit: str) -> None:
    with EnvironmentContext() as env:
        add_stage_variable("time_unit", unit)
        base = UpstageBase()
        assert base.pretty_now == f"[0.000 {unit}]"
        add_stage_variable("daily_time_count", 100)
        assert base.pretty_now == f"[Day    0 - 0.000 {unit}]"
        env.run(until=223)
        assert base.pretty_now == f"[Day    2 - 23.000 {unit}]"



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

from typing import Any

import upstage_des.api as UP
from upstage_des.type_help import TASK_GEN


class Dummy(UP.Actor):
    number = UP.State[float]()
    results = UP.State[int](default=0)


class Example(UP.Task):
    def task(self, *, actor: Dummy) -> TASK_GEN:
        yield UP.Wait(actor.number)
        actor.number /= 2


class OtherExample(UP.Task):
    def task(self, *, actor: Dummy) -> TASK_GEN:
        actor.results += 1
        yield UP.Wait(100)

    def on_interrupt(self, *, actor: Dummy, cause: Any) -> UP.InterruptStates:
        super().on_interrupt(actor=actor, cause=cause)
        return self.INTERRUPT.RESTART


fact = UP.TaskNetworkFactory(
    "example",
    {"Runner": Example},
    {"Runner": UP.TaskLinks(default="Runner", allowed=["Runner"])},
)

fact2 = UP.TaskNetworkFactory(
    "side",
    {"Side": OtherExample},
    {"Side": UP.TaskLinks(default="Side", allowed=["Side"])},
)


def test_creation() -> None:
    with UP.EnvironmentContext() as env:
        actor = Dummy(name="example", number=10)
        nuc = UP.TaskNetworkNucleus(actor=actor)
        task_net = fact.make_network()
        actor.add_task_network(task_net)
        nuc.add_network(task_net, [])
        # start the task network
        actor.start_network_loop("example", init_task_name="Runner")
        env.run(until=18)
        assert actor.number == 1.25


def test_with_interrupt() -> None:
    with UP.EnvironmentContext() as env:
        actor = Dummy(name="example", number=10, results=0)
        nuc = UP.TaskNetworkNucleus(actor=actor)

        task_net = fact.make_network()
        actor.add_task_network(task_net)
        nuc.add_network(task_net, [])

        task_net_2 = fact2.make_network()
        actor.add_task_network(task_net_2)
        nuc.add_network(task_net_2, ["number"])

        actor.start_network_loop("example", init_task_name="Runner")
        actor.start_network_loop("side", init_task_name="Side")

        env.run(until=15)
        assert actor.results == 2



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.
"""Filter stores that allow sorting of items."""

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

from simpy import Store
from simpy.resources.base import BoundClass, Get

from ..events import Get as UPGet

__all__ = ("SortedFilterStore", "SortedFilterGet")


class _SortedFilterStoreGet(Get):
    """A getter for a custom store to retrieve items using filters and priorities.

    Request to get an *item* from the *store* matching the *filter* and
    minimizing the *sorter*. The request is triggered once there is such an
    item available in the store.

    *filter* should return ``True`` for items matching the filter criterion.
    The default function returns ``True`` for all items, which makes the
    request behave exactly like :class:`StoreGet`.

    *sorter* is a function receiving one item. It should return a value that
    is to be minimized among the filter items. The default function is to not
    sort, which makes the request behave exactly like :class:`FilterStoreGet`.

    :param resource:
    :param filter: filter function for one item
    :param sorter: sort function for one item

    :type resource:
    :type filter: function
    :type sorter: function

    """

    def __init__(
        self,
        resource: "SortedFilterStore",
        filter: Callable[[Any], bool] = lambda item: True,
        sorter: Callable[[Any], tuple[Any, ...]] | None = None,
        reverse: bool = False,
    ):
        self.filter = filter
        self.sorter = sorter
        self.reverse = reverse
        super().__init__(resource)


class SortedFilterStore(Store):
    """A store that supports the filtered and sorted retrieval of items.

    Resource with *capacity* slots for storing arbitrary objects supporting
    filtered and sorted get requests. Like the :class:`Store`, the *capacity*
    is unlimited by default and objects are put and retrieved from the store in
    a first-in first-out order.

    Get requests can be customized with a filter function to only trigger for
    items for which said filter function returns ``True``. They can further be
    customized with a sorter function that prioritizes which of the filtered
    items are to be returned. The prioritization happens through a
    minimization.

    """

    # Request to get an *item* for which *filter* returns ``True`` and for
    # which *sorter* returns the minimum value out of the store.
    if TYPE_CHECKING:

        def get(self) -> _SortedFilterStoreGet:  # type: ignore[override]
            """Request to get an *item* out of the store."""
            return _SortedFilterStoreGet(self)

    else:
        get = BoundClass(_SortedFilterStoreGet)

    def _do_get(self, event: _SortedFilterStoreGet) -> bool:  # type: ignore[override]
        min_item: Any = None
        # min_val is a tuple, in case the sorter returns an iterable
        min_val: tuple[Any, ...] | None = None
        for item in self.items:
            if event.filter(item):
                if event.sorter is not None:
                    val = event.sorter(item)
                    if min_val is None or (val > min_val if event.reverse else val < min_val):
                        min_item = item
                        min_val = val
                else:
                    min_item = item
                    break
        if min_item:
            self.items.remove(min_item)
            event.succeed(min_item)
        return True


class SortedFilterGet(UPGet):
    """A Get for a SortedFilterStore."""

    def __init__(
        self,
        get_location: SortedFilterStore,
        filter: Callable[[Any], bool] = lambda item: True,
        sorter: Callable[[Any], tuple[Any, ...]] | None = None,
        reverse: bool = False,
        rehearsal_time_to_complete: float = 0.0,
    ) -> None:
        """Create a Get request on a SortedFilterStore.

        The filter function returns a boolean (True/False for in/out of consideration).

        The sorter function must return something sortable (number, tuple, e.g.)

        Args:
            get_location (SIM.Store | SIM.Container): The place for the Get request
            filter (Callable[[Any], bool]): The function that filters items in the store.
            sorter (Callable[[Any], Any]): The function that returns values to sort an item on.
            reverse (bool, optional): Whether to reverse the sort to be ascending.
            rehearsal_time_to_complete (float, optional): _description_. Defaults to 0.0.
        """
        super().__init__(
            get_location=get_location,
            rehearsal_time_to_complete=rehearsal_time_to_complete,
            filter=filter,
            sorter=sorter,
            reverse=reverse,
        )



# Copyright (C) 2024 by the Georgia Tech Research Institute (GTRI)

# Licensed under the BSD 3-Clause License.
# See the LICENSE file in the project root for complete license terms and disclaimers.

"""Great circle calculations.

These equations were largely adapted from https://edwilliams.org/avform147.htm,
although most of them can be verified from a number of open resources around
the web. The overall algorithms have been slightly modified to support UPSTAGE.
"""

from functools import lru_cache
from math import acos, asin, atan2, cos, pi, sin, sqrt
from typing import cast

from upstage_des.data_types import GeodeticLocation


@lru_cache
def get_dist_rad(point1: GeodeticLocation, point2: GeodeticLocation) -> float:
    """Get the distance (in radians) between point1 and point2.

    :param point1: the starting point
    :param point2: the ending point
    """
    point1 = point1.to_radians()
    point2 = point2.to_radians()
    ans = 2.0 * asin(
        sqrt(
            (sin((point1.lat - point2.lat) / 2.0)) ** 2
            + cos(point1.lat) * cos(point2.lat) * (sin((point1.lon - point2.lon) / 2.0)) ** 2
        )
    )
    return cast(float, ans)


@lru_cache
def get_course_rad(point1: GeodeticLocation, point2: GeodeticLocation) -> float:
    """Get the course (in radians) between point1 and point2.

    :param point1: the starting point
    :param point2: the ending point
    """
    point1 = point1.to_radians()
    point2 = point2.to_radians()
    tcl: float

    d = get_dist_rad(point1, point2)

    if sin(point2.lon - point1.lon) < 0:
        tcl = acos((sin(point2.lat) - sin(point1.lat) * cos(d)) / (sin(d) * cos(point1.lat)))
    else:
        tcl = 2.0 * pi - acos(
            (sin(point2.lat) - sin(point1.lat) * cos(d)) / (sin(d) * cos(point1.lat))
        )

    return tcl


@lru_cache
def get_pos_from_points_and_distance(
    point1: GeodeticLocation, point2: GeodeticLocation, dist: float
) -> tuple[float, float]:
    """Get a position (lat, lon) given a starting position, ending position, and distance.

    :param point1: GeodeticLocation of start of great circle
    :param point2: GeodeticLocation of end of great circle
    :param dist: (float) distance along great circle to find third point

    returns [lat, lon]
    """
    point1 = point1.to_radians()
    point2 = point2.to_radians()
    tc = get_course_rad(point1, point2)  # course from point 1 to 2

    lat: float = asin(sin(point1.lat) * cos(dist) + cos(point1.lat) * sin(dist) * cos(tc))

    dlon = atan2(
        sin(tc) * sin(dist) * cos(point1.lat),
        cos(dist) - sin(point1.lat) * sin(lat),
    )

    lon: float = ((point1.lon - dlon + pi) % (2.0 * pi)) - pi

    return (lat, lon)


@lru_cache
def get_great_circle_points(
    point_a: GeodeticLocation,
    point_b: GeodeticLocation,
    point_d: GeodeticLocation,
    dist: float,
) -> tuple[list[tuple[float, float]], list[float]] | None:
    """Let points A and B define a great circle route and D be a third point.

    Find the points on the great circle through A and B that lie a distance d from D, if they exist.

    :param point_a: GeodeticLocation of start of great circle
    :param point_b: GeodeticLocation of end of great circle
    :param point_d: GeodeticLocation, third point of interest (the center of sphere)
    :param dist: (float) distance from third to point to find intersection on great circle (radians)
    """
    point_a = point_a.to_radians()
    point_b = point_b.to_radians()
    point_d = point_d.to_radians()
    course_ad = get_course_rad(point_a, point_d)
    course_ab = get_course_rad(point_a, point_b)

    a = course_ad - course_ab
    b = get_dist_rad(point_a, point_d)

    r = (cos(b) ** 2 + sin(b) ** 2 * cos(a) ** 2) ** (
        1 / 2
    )  # arccos(r) is the cross track distance

    atd = atan2(sin(b) * cos(a), cos(b))  # the along track distance

    dist_ab = get_dist_rad(point_a, point_b)

    if cos(dist) ** 2 > r**2:
        # no points exist
        dp = None
    else:
        # two points exist
        dp = acos(cos(dist) / r)

    if dp:
        d1 = atd - dp
        d2 = atd + dp

        # make sure we can get to first crossing
        # if second cross is negative, both points are outside the D/dist radius
        if dist_ab < d1 or d2 < 0:
            return None

        p1 = get_pos_from_points_and_distance(point_a, point_b, d1)
        p2 = get_pos_from_points_and_distance(point_a, point_b, d2)

    else:
        return None

    return [p1, p2], [d1, d2]
